"""Configuration."""

import os
from typing import Literal

from dotenv import dotenv_values
from fastapi.openapi.models import OAuthFlowPassword, OAuthFlows
from pydantic import BaseModel, ConfigDict, SecretStr, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self

from scholarag.generative_question_answering import PROMPT_TEMPLATE


class SettingsKeycloak(BaseModel):
    """Class retrieving keycloak info for authorization."""

    issuer: str | None = None
    validate_token: bool = False

    model_config = ConfigDict(frozen=True)

    @model_validator(mode="after")
    def check_issuer(self) -> Self:
        """Check if there is an issuer provided for authentication."""
        if self.validate_token and (self.issuer is None):
            raise ValueError("No issuer provided.")
        return self

    @property
    def token_endpoint(self) -> str:
        """Define the token endpoint."""
        return f"{self.issuer}/protocol/openid-connect/token"

    @property
    def user_info_endpoint(self) -> str | None:
        """Define the user_info endpoint."""
        if self.validate_token:
            return f"{self.issuer}/protocol/openid-connect/userinfo"
        else:
            return None

    @property
    def flows(self) -> OAuthFlows:
        """Define the flow to override Fastapi's one."""
        return OAuthFlows(
            password=OAuthFlowPassword(
                tokenUrl=self.token_endpoint,
            ),
        )


class SettingsDB(BaseModel):
    """Database settings."""

    db_type: Literal["opensearch", "elasticsearch"]
    index_paragraphs: str
    host: str
    port: int
    index_journals: str | None = None
    user: str | None = None
    password: SecretStr | None = None

    model_config = ConfigDict(frozen=True)


class SettingsRetrieval(BaseModel):
    """Retrieval settings."""

    max_length: int = 100000

    model_config = ConfigDict(frozen=True)


class SettingsCohereReranking(BaseModel):
    """Settings cohere reranker."""

    cohere_token: SecretStr | None = None

    model_config = ConfigDict(frozen=True)


class SettingsOpenAI(BaseModel):
    """OpenAI settings."""

    token: SecretStr | None = None
    model: str = "gpt-3.5-turbo"
    temperature: float = 0
    max_tokens: int | None = None

    model_config = ConfigDict(frozen=True)


class SettingsGenerative(BaseModel):
    """Generative QA settings."""

    openai: SettingsOpenAI = SettingsOpenAI()
    prompt_template: SecretStr = SecretStr(PROMPT_TEMPLATE)

    model_config = ConfigDict(frozen=True)


class SettingsRedisCaching(BaseModel):
    """Redis settings."""

    host: str | None = None
    port: int | None = None
    expiry: float = 30.0  # Time in days

    model_config = ConfigDict(frozen=True)


class SettingsMetadata(BaseModel):
    """Metadata settings."""

    external_apis: bool = True
    timeout: int = 30

    model_config = ConfigDict(frozen=True)


class SettingsLogging(BaseModel):
    """Metadata settings."""

    level: Literal["debug", "info", "warning", "error", "critical"] = "info"
    external_packages: Literal["debug", "info", "warning", "error", "critical"] = (
        "warning"
    )

    model_config = ConfigDict(frozen=True)


class SettingsMisc(BaseModel):
    """Other settings."""

    application_prefix: str = ""
    # list is not hashable, the cors_origins have to be provided as a string with
    # comma separated entries, i.e. "value_1, value_2, ..."
    cors_origins: str = ""

    model_config = ConfigDict(frozen=True)


class Settings(BaseSettings):
    """All settings."""

    db: SettingsDB
    retrieval: SettingsRetrieval = SettingsRetrieval()  # has no required
    reranking: SettingsCohereReranking = SettingsCohereReranking()  # has no required
    generative: SettingsGenerative = SettingsGenerative()  # has no required
    redis: SettingsRedisCaching = SettingsRedisCaching()  # has no required
    metadata: SettingsMetadata = SettingsMetadata()  # has no required
    logging: SettingsLogging = SettingsLogging()  # has no required
    keycloak: SettingsKeycloak = SettingsKeycloak()  # has no required
    misc: SettingsMisc = SettingsMisc()  # has no required

    model_config = SettingsConfigDict(
        env_file=".env",
        env_prefix="SCHOLARAG__",
        env_nested_delimiter="__",
        frozen=True,
    )


# Load the remaining variables into the environment
# Necessary for things like SSL_CERT_FILE
config = dotenv_values()
for k, v in config.items():
    if k.lower().startswith("scholarag_"):
        continue
    if v is None:
        continue
    os.environ[k] = os.environ.get(k, v)  # environment has precedence



"""Document store class connecting to Opensearch."""

import logging
from collections.abc import AsyncGenerator, Iterable
from typing import Any

from opensearchpy import AsyncOpenSearch as AsyncOpensearch
from opensearchpy import OpenSearch as Opensearch
from opensearchpy.helpers import async_bulk, async_scan, bulk, scan
from pydantic import model_validator

from scholarag.document_stores import AsyncBaseSearch, BaseSearch

logger = logging.getLogger(__name__)

SETTINGS: dict[str, Any] = {
    "number_of_shards": 2,
    "number_of_replicas": 1,
}

MAPPINGS_PARAGRAPHS: dict[str, Any] = {
    "dynamic": "true",
    "properties": {
        "article_id": {"type": "keyword"},
        "doi": {"type": "keyword"},
        "pmc_id": {"type": "keyword"},
        "pubmed_id": {"type": "keyword"},
        "arxiv_id": {"type": "keyword"},
        "title": {"type": "text"},
        "authors": {"type": "keyword"},
        "journal": {"type": "keyword"},
        "date": {"type": "date", "format": "yyyy-MM-dd"},
        "paragraph_id": {"type": "short"},
        "text": {"type": "text"},
        "article_type": {"type": "keyword"},
        "section": {"type": "keyword"},
    },
}


class OpenSearch(BaseSearch):
    """Class to use sync Opensearch as document store."""

    client: Opensearch

    @model_validator(mode="before")
    def connect_to_ds(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Connect to OS."""
        values["use_ssl_and_verify_certs"] = values.get(
            "use_ssl_and_verify_certs", True
        )  # by default we use SSL

        if values["host"].startswith("http"):
            values["host"] = values["host"].rpartition("://")[-1]
        if values.get("user") is None and values.get("password") is None:
            client = Opensearch(
                hosts=[{"host": values["host"], "port": values["port"]}],
                verify_certs=values["use_ssl_and_verify_certs"],
                use_ssl=values["use_ssl_and_verify_certs"],
                request_timeout=60,
                max_retries=3,
                retry_on_timeout=True,
                timeout=60,
            )
        else:
            client = Opensearch(
                hosts=[{"host": values["host"], "port": values["port"]}],
                http_auth=(values["user"], values["password"]),
                verify_certs=values["use_ssl_and_verify_certs"],
                use_ssl=values["use_ssl_and_verify_certs"],
                request_timeout=60,
                max_retries=3,
                retry_on_timeout=True,
                timout=60,
            )

        values["client"] = client
        return values

    def get_available_indexes(self) -> list[str]:
        """Return all available indexes."""
        return list(self.client.indices.get_alias().keys())

    def remove_index(self, index: str) -> None:
        """Remove an index."""
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        self.client.indices.delete(index=index)
        logger.info(f"Index {index} deleted successfully")

    def get_document(self, index: str, doc_id: str) -> dict[str, Any]:
        """Return a document.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_id
            ID under which the document is indexed.

        Returns
        -------
        doc
            Document retrieved.
        """
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        if not self.client.exists(index=index, id=doc_id):
            raise RuntimeError("Document not in the document store")
        return self.client.get(index=index, id=doc_id)["_source"]

    def get_documents(self, index: str, doc_ids: list[str]) -> list[dict[str, Any]]:
        """Return a list of documents.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_ids
            list of ids under which the documents are indexed.

        Returns
        -------
        docs
            list of document retrieved.
        """
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        docs = self.client.mget(index=index, body={"ids": doc_ids})
        docs_to_return = [
            {"document_id": doc["_id"], **doc["_source"]}
            for doc in docs["docs"]
            if doc["found"]
        ]
        return docs_to_return

    @staticmethod
    def _process_search_hits(res: Any) -> list[dict[str, Any]]:
        """Process search hits.

        Parameters
        ----------
        res
            Result of a DB query.

        Returns
        -------
        out
            Processed results.
        """
        out = []
        for hit in res["hits"]["hits"]:
            row = {}

            row["article_id"] = hit["_source"].get("article_id")
            row["title"] = hit["_source"].get("title")
            row["authors"] = hit["_source"].get("authors")
            row["doi"] = hit["_source"].get("doi")
            row["pubmed_id"] = hit["_source"].get("pubmed_id")
            row["section"] = hit["_source"].get("section")
            row["document_id"] = hit["_id"]
            row["date"] = hit["_source"].get("date")
            row["journal"] = hit["_source"].get("journal")
            row["paragraph_id"] = hit["_source"].get("paragraph_id")
            row["pmc_id"] = hit["_source"].get("pmc_id")
            row["arxiv_id"] = hit["_source"].get("arxiv_id")
            row["text"] = hit["_source"].get("text")
            row["article_type"] = hit["_source"].get("article_type")
            row["score"] = hit["_score"]

            out.append(row)

        return out

    def get_index_mappings(self, index: str) -> dict[str, Any]:
        """Return an index mapping."""
        return self.client.indices.get_mapping(index=index)[index]["mappings"]

    def create_index(
        self,
        index: str,
        settings: dict[str, Any] | None,
        mappings: dict[str, Any] | None,
    ) -> None:
        """Create a new index."""
        if index in self.get_available_indexes():
            raise RuntimeError("Index already in OS")
        body = {"settings": settings, "mappings": mappings}
        self.client.indices.create(index=index, body=body)
        logger.info(f"Index {index} created successfully")

    def add_fields(
        self,
        index: str,
        settings: dict[str, Any] | None = None,
        mapping: dict[str, Any] | None = None,
    ) -> None:
        """Update the index with a new mapping and settings."""
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in OS")

        if settings:
            self.client.indices.put_settings(index=index, body=settings)
        if mapping:
            current_mapping = self.get_index_mappings(index=index)
            existing_properties = current_mapping["properties"]
            updated_mapping = {"properties": {**existing_properties, **mapping}}
            self.client.indices.put_mapping(index=index, body=updated_mapping)
        logger.info(f"Index {index} updated successfully.")

    def count_documents(self, index: str, query: dict[str, Any] | None = None) -> int:
        """Return the number of documents in an index.

        Parameters
        ----------
        index
            OS index where documents are stored.
        query
            Optional query to filter the documents to count.

        Returns
        -------
        document_count
            Number of documents in index that respect query.
        """
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in OS")

        return self.client.count(index=index, body=query)["count"]

    def exists(self, index: str, doc_id: str) -> bool:
        """Return True if this document exists in the index.

        Parameters
        ----------
        index
            OS index where documents are stored.
        doc_id
            ID under which the document might be indexed.

        Returns
        -------
            True if the document exists within an index.
        """
        return self.client.exists(index=index, id=doc_id)

    def iter_document(
        self, index: str, query: dict[str, Any] | None = None, size: int = 1000
    ) -> Iterable[dict[str, Any]]:
        """Scan the documents matching a specific query."""
        if query is None:
            query = {"query": {"match_all": {}}}
        doc_gen = scan(self.client, index=index, query=query, size=size)
        return doc_gen

    def add_document(
        self, index: str, doc: dict[str, Any], doc_id: str | None = None
    ) -> None:
        """Index a document.

        Parameters
        ----------
        index
            OS index where documents are stored.
        doc
            Document to index.
        doc_id
            ID under which the document must be indexed.
        """
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in OS")
        if doc_id and self.client.exists(index=index, id=doc_id):
            raise RuntimeError("Document already in OS")
        self.client.index(index=index, id=doc_id, body=doc)
        logger.info(f"Document {doc_id} indexed successfully.")

    def bulk(
        self,
        actions: list[dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Bulk upload of documents."""
        bulk(client=self.client, actions=actions, **kwargs)
        logger.info("Successfully updated documents in bulk")

    def search(
        self,
        index: str,
        query: dict[str, Any] | None = None,
        size: int = 10,
        aggs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Wrap around the search api."""
        body = None
        if query is not None:
            if "query" not in query.keys():
                body = {"query": query}
            else:
                body = query
        if aggs is not None:
            if body is not None:
                body["aggs"] = aggs
            else:
                body = {"aggs": aggs}
        return self.client.search(index=index, body=body, size=size, **kwargs)

    def bm25_search(
        self,
        index_doc: str,
        query: str,
        filter_query: dict[str, Any] | None = None,
        k: int = 10,
    ) -> list[dict[str, Any]]:
        """BM25 search.

        Parameters
        ----------
        index_doc
            OS index where documents are stored.
        query
            Query to retrieve documents that match it.
        k
            Number of documents to return

        Returns
        -------
        res
            Documents retrieved.
        """
        if index_doc not in self.get_available_indexes():
            raise RuntimeError("Index for documents not in OS")

        if filter_query is not None:
            query_dict = {
                "query": {
                    "bool": {"must": {"match": {"text": query}}, "filter": filter_query}
                }
            }
        else:
            query_dict = {"query": {"bool": {"must": {"match": {"text": query}}}}}

        res = self.client.search(
            index=index_doc,
            body=query_dict,
            size=k,
        )

        out = self._process_search_hits(res)
        return out


class AsyncOpenSearch(AsyncBaseSearch):
    """Class to use async Opensearch as document store."""

    client: AsyncOpensearch

    @model_validator(mode="before")
    def connect_to_ds(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Connect to OS."""
        values["use_ssl_and_verify_certs"] = values.get(
            "use_ssl_and_verify_certs", True
        )  # by default we use SSL

        if values["host"].startswith("http"):
            values["host"] = values["host"].rpartition("://")[-1]
        if values.get("user") is None and values.get("password") is None:
            client = AsyncOpensearch(
                hosts=[{"host": values["host"], "port": values["port"]}],
                verify_certs=values["use_ssl_and_verify_certs"],
                use_ssl=values["use_ssl_and_verify_certs"],
                request_timeout=60,
                max_retries=3,
                retry_on_timeout=True,
                timeout=60,
            )
        else:
            client = AsyncOpensearch(
                hosts=[{"host": values["host"], "port": values["port"]}],
                http_auth=(values["user"], values["password"]),
                verify_certs=values["use_ssl_and_verify_certs"],
                use_ssl=values["use_ssl_and_verify_certs"],
                request_timeout=60,
                max_retries=3,
                retry_on_timeout=True,
                timeout=60,
            )

        values["client"] = client
        return values

    async def get_available_indexes(self) -> list[str]:
        """Return all available indexes."""
        indices = await self.client.indices.get_alias()
        return list(indices.keys())

    async def remove_index(self, index: str) -> None:
        """Remove an index."""
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in document store")
        await self.client.indices.delete(index=index)
        logger.info(f"Index {index} deleted successfully")

    async def get_document(self, index: str, doc_id: str) -> dict[str, Any]:
        """Return a document.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_id
            ID under which the document is indexed.

        Returns
        -------
        doc
            Document retrieved.
        """
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        if not await self.client.exists(index=index, id=doc_id):
            raise RuntimeError("Document not in the document store")
        doc = await self.client.get(index=index, id=doc_id)
        return doc["_source"]

    async def get_documents(
        self, index: str, doc_ids: list[str]
    ) -> list[dict[str, Any]]:
        """Return a list of documents.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_ids
            list of ids under which the documents are indexed.

        Returns
        -------
        docs
            list of document retrieved.
        """
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        docs = await self.client.mget(index=index, body={"ids": doc_ids})
        docs_to_return = [
            {"document_id": doc["_id"], **doc["_source"]}
            for doc in docs["docs"]
            if doc["found"]
        ]
        return docs_to_return

    @staticmethod
    def _process_search_hits(res: Any) -> list[dict[str, Any]]:
        """Process search hits.

        Parameters
        ----------
        res
            Result of a DB query.

        Returns
        -------
        out
            Processed results.
        """
        out = []
        for hit in res["hits"]["hits"]:
            row = {}

            row["article_id"] = hit["_source"].get("article_id")
            row["title"] = hit["_source"].get("title")
            row["authors"] = hit["_source"].get("authors")
            row["doi"] = hit["_source"].get("doi")
            row["pubmed_id"] = hit["_source"].get("pubmed_id")
            row["section"] = hit["_source"].get("section")
            row["document_id"] = hit["_id"]
            row["date"] = hit["_source"].get("date")
            row["journal"] = hit["_source"].get("journal")
            row["paragraph_id"] = hit["_source"].get("paragraph_id")
            row["pmc_id"] = hit["_source"].get("pmc_id")
            row["arxiv_id"] = hit["_source"].get("arxiv_id")
            row["text"] = hit["_source"].get("text")
            row["article_type"] = hit["_source"].get("article_type")
            row["score"] = hit["_score"]

            out.append(row)

        return out

    async def get_index_mappings(self, index: str) -> dict[str, Any]:
        """Return an index mapping."""
        mapping = await self.client.indices.get_mapping(index=index)
        return mapping[index]["mappings"]

    async def create_index(
        self,
        index: str,
        settings: dict[str, Any] | None,
        mappings: dict[str, Any] | None,
    ) -> None:
        """Create a new index."""
        if index in await self.get_available_indexes():
            raise RuntimeError("Index already in OS")
        body = {"settings": settings, "mappings": mappings}
        await self.client.indices.create(index=index, body=body)
        logger.info(f"Index {index} created successfully")

    async def add_fields(
        self,
        index: str,
        settings: dict[str, Any] | None = None,
        mapping: dict[str, Any] | None = None,
    ) -> None:
        """Update the index with a new mapping and settings."""
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in OS")

        if settings:
            await self.client.indices.put_settings(index=index, body=settings)
        if mapping:
            current_mapping = await self.get_index_mappings(index=index)
            existing_properties = current_mapping["properties"]
            updated_mapping = {"properties": {**existing_properties, **mapping}}
            await self.client.indices.put_mapping(index=index, body=updated_mapping)
        logger.info(f"Index {index} updated successfully.")

    async def count_documents(
        self, index: str, query: dict[str, Any] | None = None
    ) -> int:
        """Return the number of documents in an index.

        Parameters
        ----------
        index
            OS index where documents are stored.
        query
            Optional query to filter the documents to count.

        Returns
        -------
        document_count
            Number of documents in index that respect query.
        """
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in OS")
        count = await self.client.count(index=index, body=query)
        return count["count"]

    async def exists(self, index: str, doc_id: str) -> bool:
        """Return True if this document exists in the index.

        Parameters
        ----------
        index
            OS index where documents are stored.
        doc_id
            ID under which the document might be indexed.

        Returns
        -------
            True if the document exists within an index.
        """
        return await self.client.exists(index=index, id=doc_id)

    def iter_document(
        self, index: str, query: dict[str, Any] | None = None, size: int = 1000
    ) -> AsyncGenerator[int, None]:
        """Scan the documents matching a specific query."""
        if query is None:
            query = {"query": {"match_all": {}}}
        doc_gen = async_scan(self.client, index=index, query=query, size=size)
        return doc_gen

    async def add_document(
        self, index: str, doc: dict[str, Any], doc_id: str | None = None
    ) -> None:
        """Index a document.

        Parameters
        ----------
        index
            OS index where documents are stored.
        doc
            Document to index.
        doc_id
            ID under which the document must be indexed.
        """
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in OS")
        if doc_id and await self.client.exists(index=index, id=doc_id):
            raise RuntimeError("Document already in OS")
        await self.client.index(index=index, id=doc_id, body=doc)
        logger.info(f"Document {doc_id} indexed successfully.")

    async def bulk(
        self,
        actions: list[dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Bulk upload of documents."""
        await async_bulk(client=self.client, actions=actions, **kwargs)
        logger.info("Successfully updated documents in bulk")

    async def search(
        self,
        index: str,
        query: dict[str, Any] | None = None,
        size: int = 10,
        aggs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Wrap around the search api."""
        body = None
        if query is not None:
            if "query" not in query.keys():
                body = {"query": query}
            else:
                body = query
        if aggs is not None:
            if body is not None:
                body["aggs"] = aggs
            else:
                body = {"aggs": aggs}
        return await self.client.search(index=index, body=body, size=size, **kwargs)

    async def bm25_search(
        self,
        index_doc: str,
        query: str | dict[str, Any],
        filter_query: dict[str, Any] | None = None,
        k: int = 10,
    ) -> list[dict[str, Any]]:
        """BM25 search.

        Parameters
        ----------
        index_doc
            OS index where documents are stored.
        query
            Query to retrieve documents that match it.
        filter_query
            Filtering parameters for the query
        k
            Number of documents to return

        Returns
        -------
        res
            Documents retrieved.
        """
        if index_doc not in await self.get_available_indexes():
            raise RuntimeError("Index for documents not in OS")

        if filter_query is not None:
            query_dict = {
                "query": {
                    "bool": {"must": {"match": {"text": query}}, "filter": filter_query}
                }
            }
        else:
            query_dict = {"query": {"bool": {"must": {"match": {"text": query}}}}}

        res = await self.client.search(
            index=index_doc,
            body=query_dict,
            size=k,
        )

        out = self._process_search_hits(res)
        return out

    async def close(self) -> None:
        """Close the aiohttp session."""
        await self.client.close()



"""Document store class connecting to Elasticsearch."""

import logging
from collections.abc import AsyncIterable, Iterable
from typing import Any

from elasticsearch import AsyncElasticsearch, Elasticsearch
from elasticsearch.helpers import async_bulk, async_scan, bulk, scan
from pydantic import model_validator

from scholarag.document_stores import AsyncBaseSearch, BaseSearch

logger = logging.getLogger(__name__)

SETTINGS: dict[str, Any] = {"number_of_shards": 2, "number_of_replicas": 1}

MAPPINGS_PARAGRAPHS: dict[str, Any] = {
    "dynamic": "strict",
    "properties": {
        "article_id": {"type": "keyword"},
        "doi": {"type": "keyword"},
        "pmc_id": {"type": "keyword"},
        "pubmed_id": {"type": "keyword"},
        "arxiv_id": {"type": "keyword"},
        "title": {"type": "text"},
        "authors": {"type": "keyword"},
        "journal": {"type": "keyword"},
        "date": {"type": "date", "format": "yyyy-MM-dd"},
        "section": {"type": "keyword"},
        "paragraph_id": {"type": "short"},
        "text": {"type": "text"},
        "article_type": {"type": "keyword"},
    },
}


class ElasticSearch(BaseSearch):
    """Class to use sync Elasticsearch as document store."""

    client: Elasticsearch

    @model_validator(mode="before")
    def connect_to_ds(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Connect to ES."""
        values["use_ssl_and_verify_certs"] = values.get(
            "use_ssl_and_verify_certs", False
        )  # by default we don't use SSL
        if values.get("user") is None and values.get("password") is None:
            client = Elasticsearch(
                f"{values['host']}:{values['port']}",
                verify_certs=values["use_ssl_and_verify_certs"],
                request_timeout=60,
                max_retries=3,
                retry_on_timeout=True,
            )
        else:
            client = Elasticsearch(
                f"{values['host']}:{values['port']}",
                basic_auth=(values["user"], values["password"]),
                verify_certs=values["use_ssl_and_verify_certs"],
                request_timeout=60,
                max_retries=3,
                retry_on_timeout=True,
            )

        values["client"] = client
        return values

    def get_available_indexes(self) -> list[str]:
        """Return all available indexes."""
        return list(self.client.indices.get_alias().keys())

    def remove_index(self, index: str) -> None:
        """Remove an index."""
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        self.client.indices.delete(index=index)
        logger.info(f"Index {index} deleted successfully")

    def get_document(self, index: str, doc_id: str) -> dict[str, Any]:
        """Return a document.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_id
            ID under which the document is indexed.

        Returns
        -------
        doc
            Document retrieved.
        """
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        if not self.client.exists(index=index, id=doc_id):
            raise RuntimeError("Document not in the document store")
        return self.client.get(index=index, id=doc_id)["_source"]

    def get_documents(self, index: str, doc_ids: list[str]) -> list[dict[str, Any]]:
        """Return a list of documents.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_ids
            list of ids under which the documents are indexed.

        Returns
        -------
        docs
            list of document retrieved.
        """
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        docs = self.client.mget(index=index, ids=doc_ids)
        docs_to_return = [
            {"document_id": doc["_id"], **doc["_source"]}
            for doc in docs["docs"]
            if doc["found"]
        ]
        return docs_to_return

    @staticmethod
    def _process_search_hits(res: Any) -> list[dict[str, Any]]:
        """Process search hits.

        Parameters
        ----------
        res
            Result of a DB query.

        Returns
        -------
        out
            Processed results.
        """
        out = []
        for hit in res["hits"]["hits"]:
            row = {}

            row["article_id"] = hit["_source"].get("article_id")
            row["title"] = hit["_source"].get("title")
            row["authors"] = hit["_source"].get("authors")
            row["doi"] = hit["_source"].get("doi")
            row["pubmed_id"] = hit["_source"].get("pubmed_id")
            row["section"] = hit["_source"].get("section")
            row["date"] = hit["_source"].get("date")
            row["journal"] = hit["_source"].get("journal")
            row["document_id"] = hit["_id"]
            row["paragraph_id"] = hit["_source"].get("paragraph_id")
            row["pmc_id"] = hit["_source"].get("pmc_id")
            row["arxiv_id"] = hit["_source"].get("arxiv_id")
            row["text"] = hit["_source"].get("text")
            row["article_type"] = hit["_source"].get("article_type")
            row["score"] = hit["_score"]

            out.append(row)

        return out

    def get_index_mappings(self, index: str) -> dict[str, Any]:
        """Return an index mapping."""
        return self.client.indices.get_mapping(index=index).raw[index]["mappings"]

    def create_index(
        self,
        index: str,
        settings: dict[str, Any] | None,
        mappings: dict[str, Any] | None,
    ) -> None:
        """Create a new index."""
        if index in self.get_available_indexes():
            raise RuntimeError("Index already in ES")
        self.client.indices.create(index=index, settings=settings, mappings=mappings)
        logger.info(f"Index {index} created successfully")

    def add_fields(
        self,
        index: str,
        settings: dict[str, Any] | None = None,
        mapping: dict[str, Any] | None = None,
    ) -> None:
        """Update the index with a new mapping and settings."""
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in ES")
        if settings:
            self.client.indices.put_settings(index=index, settings=settings)
        if mapping:
            self.client.indices.put_mapping(index=index, properties=mapping)
        logger.info(f"Index {index} updated successfully.")

    def count_documents(self, index: str, query: dict[str, Any] | None = None) -> int:
        """Return the number of documents in an index.

        Parameters
        ----------
        index
            ES index where documents are stored.
        query
            Optional query to filter the documents to count.

        Returns
        -------
        document_count
            Number of documents in index that respect query.
        """
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in ES")
        if query is None:
            query = {"match_all": {}}

        return self.client.count(index=index, query=query)["count"]

    def exists(self, index: str, doc_id: str) -> bool:
        """Return True if this document exists in the index.

        Parameters
        ----------
        index
            ES index where documents are stored.
        doc_id
            ID under which the document might be indexed.

        Returns
        -------
            True if the document exists within an index.
        """
        return bool(self.client.exists(index=index, id=doc_id))

    def iter_document(
        self, index: str, query: dict[str, Any] | None = None, size: int = 1000
    ) -> Iterable[dict[str, Any]]:
        """Scan the documents matching a specific query."""
        if query is None:
            query = {"query": {"match_all": {}}}
        doc_gen = scan(self.client, index=index, query=query, size=size)
        return doc_gen

    def add_document(
        self, index: str, doc: dict[str, Any], doc_id: str | None = None
    ) -> None:
        """Index a document.

        Parameters
        ----------
        index
            ES index where documents are stored.
        doc
            Document to index.
        doc_id
            ID under which the document must be indexed.
        """
        if index not in self.get_available_indexes():
            raise RuntimeError("Index not in ES")
        if doc_id and self.client.exists(index=index, id=doc_id):
            raise RuntimeError("Document already in ES")
        self.client.index(index=index, id=doc_id, document=doc)
        logger.info(f"Document {doc_id} indexed successfully.")

    def bulk(
        self,
        actions: list[dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Bulk upload of documents."""
        bulk(client=self.client, actions=actions, **kwargs)
        logger.info("Successfully updated documents in bulk")

    def search(
        self,
        index: str,
        query: dict[str, Any] | None = None,
        size: int = 10,
        aggs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Wrap around the search api."""
        if query is not None and "query" in query.keys():
            query = query["query"]
        query = postprocess_query(query)
        return self.client.search(
            index=index, query=query, size=size, aggs=aggs, **kwargs
        ).raw

    def bm25_search(
        self,
        index_doc: str,
        query: str,
        filter_query: dict[str, Any] | None = None,
        k: int = 10,
    ) -> list[dict[str, Any]]:
        """BM25 search.

        Parameters
        ----------
        index_doc
            ES index where documents are stored.
        query
            Query to retrieve documents that match it.
        filter_query
            Dictionary containing filtering options.
        k
            Number of documents to return

        Returns
        -------
        res
            Documents retrieved.
        """
        if index_doc not in self.get_available_indexes():
            raise RuntimeError("Index for documents not in ES")
        if filter_query is not None:
            query_dict = {
                "bool": {"must": {"match": {"text": query}}, "filter": filter_query}
            }
        else:
            query_dict = {"bool": {"must": {"match": {"text": query}}}}

        res = self.client.search(
            index=index_doc,
            query=query_dict,
            size=k,
        )

        out = self._process_search_hits(res)
        return out


class AsyncElasticSearch(AsyncBaseSearch):
    """Class to use async Elasticsearch."""

    client: AsyncElasticsearch

    @model_validator(mode="before")
    def connect_to_ds(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Connect to ES."""
        values["use_ssl_and_verify_certs"] = values.get(
            "use_ssl_and_verify_certs", False
        )  # by default we don't use SSL
        if values.get("user") is None and values.get("password") is None:
            client = AsyncElasticsearch(
                f"{values['host']}:{values['port']}",
                verify_certs=values["use_ssl_and_verify_certs"],
                request_timeout=60,
                max_retries=3,
                retry_on_timeout=True,
            )
        else:
            client = AsyncElasticsearch(
                f"{values['host']}:{values['port']}",
                basic_auth=(values["user"], values["password"]),
                verify_certs=values["use_ssl_and_verify_certs"],
                request_timeout=60,
                max_retries=3,
                retry_on_timeout=True,
            )

        values["client"] = client
        return values

    async def get_available_indexes(self) -> list[str]:
        """Return all available indexes."""
        indices = await self.client.indices.get_alias()
        return list(indices.keys())

    async def remove_index(self, index: str) -> None:
        """Remove an index."""
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in document store")
        await self.client.indices.delete(index=index)
        logger.info(f"Index {index} deleted successfully")

    async def get_document(self, index: str, doc_id: str) -> dict[str, Any]:
        """Return a document.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_id
            ID under which the document is indexed.

        Returns
        -------
        doc
            Document retrieved.
        """
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        if not await self.client.exists(index=index, id=doc_id):
            raise RuntimeError("Document not in the document store")
        doc = await self.client.get(index=index, id=doc_id)
        return doc["_source"]

    async def get_documents(
        self, index: str, doc_ids: list[str]
    ) -> list[dict[str, Any]]:
        """Return a list of documents.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_ids
            list of ids under which the documents are indexed.

        Returns
        -------
        docs
            list of document retrieved.
        """
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in the document store")
        docs = await self.client.mget(index=index, ids=doc_ids)
        docs_to_return = [
            {"document_id": doc["_id"], **doc["_source"]}
            for doc in docs["docs"]
            if doc["found"]
        ]
        return docs_to_return

    @staticmethod
    def _process_search_hits(res: Any) -> list[dict[str, Any]]:
        """Process search hits.

        Parameters
        ----------
        res
            Result of a DB query.

        Returns
        -------
        out
            Processed results.
        """
        out = []
        for hit in res["hits"]["hits"]:
            row = {}

            row["article_id"] = hit["_source"].get("article_id")
            row["title"] = hit["_source"].get("title")
            row["authors"] = hit["_source"].get("authors")
            row["doi"] = hit["_source"].get("doi")
            row["pubmed_id"] = hit["_source"].get("pubmed_id")
            row["pmc_id"] = hit["_source"].get("pmc_id")
            row["arxiv_id"] = hit["_source"].get("arxiv_id")
            row["section"] = hit["_source"].get("section")
            row["document_id"] = hit["_id"]
            row["date"] = hit["_source"].get("date")
            row["journal"] = hit["_source"].get("journal")
            row["paragraph_id"] = hit["_source"].get("paragraph_id")
            row["text"] = hit["_source"].get("text")
            row["article_type"] = hit["_source"].get("article_type")
            row["score"] = hit["_score"]

            out.append(row)

        return out

    async def get_index_mappings(self, index: str) -> dict[str, Any]:
        """Return an index mapping."""
        mapping = await self.client.indices.get_mapping(index=index)
        return mapping.raw[index]["mappings"]

    async def create_index(
        self,
        index: str,
        settings: dict[str, Any] | None,
        mappings: dict[str, Any] | None,
    ) -> None:
        """Create a new index."""
        if index in await self.get_available_indexes():
            raise RuntimeError("Index already in ES")
        await self.client.indices.create(
            index=index, settings=settings, mappings=mappings
        )
        logger.info(f"Index {index} created successfully")

    async def add_fields(
        self,
        index: str,
        settings: dict[str, Any] | None = None,
        mapping: dict[str, Any] | None = None,
    ) -> None:
        """Update the index with a new mapping and settings."""
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in ES")

        if settings:
            await self.client.indices.put_settings(index=index, settings=settings)
        if mapping:
            await self.client.indices.put_mapping(index=index, properties=mapping)
        logger.info(f"Index {index} updated successfully.")

    async def count_documents(
        self, index: str, query: dict[str, Any] | None = None
    ) -> int:
        """Return the number of documents in an index.

        Parameters
        ----------
        index
            ES index where documents are stored.
        query
            Optional query to filter the documents to count.

        Returns
        -------
        document_count
            Number of documents in index that respect query.
        """
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in ES")
        if query is None:
            query = {"match_all": {}}

        count = await self.client.count(index=index, query=query)

        return count["count"]

    async def exists(self, index: str, doc_id: str) -> bool:
        """Return True if this document exists in the index.

        Parameters
        ----------
        index
            ES index where documents are stored.
        doc_id
            ID under which the document might be indexed.

        Returns
        -------
            True if the document exists within an index.
        """
        return bool(await self.client.exists(index=index, id=doc_id))

    def iter_document(
        self, index: str, query: dict[str, Any] | None = None, size: int = 1000
    ) -> AsyncIterable[dict[str, Any]]:
        """Scan the documents matching a specific query."""
        if query is None:
            query = {"query": {"match_all": {}}}
        doc_gen = async_scan(self.client, index=index, query=query, size=size)
        return doc_gen

    async def add_document(
        self, index: str, doc: dict[str, Any], doc_id: str | None = None
    ) -> None:
        """Index a document.

        Parameters
        ----------
        index
            ES index where documents are stored.
        doc
            Document to index.
        doc_id
            ID under which the document must be indexed.
        """
        if index not in await self.get_available_indexes():
            raise RuntimeError("Index not in ES")
        if doc_id and await self.client.exists(index=index, id=doc_id):
            raise RuntimeError("Document already in ES")
        await self.client.index(index=index, id=doc_id, document=doc)
        logger.info(f"Document {doc_id} indexed successfully.")

    async def bulk(
        self,
        actions: list[dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Bulk upload of documents."""
        await async_bulk(client=self.client, actions=actions, **kwargs)
        logger.info("Successfully updated documents in bulk")

    async def search(
        self,
        index: str,
        query: dict[str, Any] | None = None,
        size: int = 10,
        aggs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Wrap around the search api."""
        if query is not None and "query" in query.keys():
            query = query["query"]
        query = postprocess_query(query)
        res = await self.client.search(
            index=index, query=query, size=size, aggs=aggs, **kwargs
        )
        return res.raw

    async def bm25_search(
        self,
        index_doc: str,
        query: str | dict[str, Any],
        filter_query: dict[str, Any] | None = None,
        k: int = 10,
    ) -> list[dict[str, Any]]:
        """BM25 search.

        Parameters
        ----------
        index_doc
            ES index where documents are stored.
        query
            Query to retrieve documents that match it.
        filter_query
            Dictionary of filtering criteria to add to the query.
        k
            Number of documents to return

        Returns
        -------
        res
            Documents retrieved.
        """
        if index_doc not in await self.get_available_indexes():
            raise RuntimeError("Index for documents not in ES")

        if filter_query is not None:
            query_dict = {
                "bool": {"must": {"match": {"text": query}}, "filter": filter_query}
            }
        else:
            query_dict = {"bool": {"must": {"match": {"text": query}}}}

        res = await self.client.search(
            index=index_doc,
            query=query_dict,
            size=k,
        )

        out = self._process_search_hits(res)
        return out

    async def close(self) -> None:
        """Close the aiohttp session."""
        await self.client.close()


def postprocess_query(query: dict[str, Any] | None) -> dict[str, Any] | None:
    """Post-process the query to handle special characters."""
    if query is None:
        return None

    if "query_string" in query.keys() and "query" in query["query_string"]:
        special_characters = {
            "+",
            "-",
            "=",
            "&&",
            "||",
            "!",
            "{",
            "}",
            "[",
            "]",
            "^",
            "~",
            "*",
            "?",
            ":",
            r"\"",
            "/",
            ">",
            "<",
        }
        intersection = set(query["query_string"]["query"]).intersection(
            special_characters
        )
        if len(intersection) > 0:
            query_str = query["query_string"]["query"]
            for character in intersection:
                if character in {">", "<"}:
                    query_str = query_str.replace(">", "")
                    query_str = query_str.replace("<", "")
                else:
                    query_str = query_str.replace(character, "\\" + character)
            query["query_string"]["query"] = query_str
            logger.info(f"Query had to be slightly modified to: {query}")
    return query



import json
import unittest.mock as mock
from unittest.mock import MagicMock

import httpx
import pytest
from scholarag.document_stores import AsyncBaseSearch
from scholarag.retrieve_metadata import (
    MetaDataRetriever,
    get_citation_count,
    get_impact_factors,
    get_journal_name,
    recreate_abstract,
)


class AsyncIterator:
    def __init__(self, seq):
        self.iter = iter(seq)

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            return next(self.iter)
        except StopIteration:
            raise StopAsyncIteration


@pytest.mark.parametrize(
    "issns,return_value,response",
    [
        (
            ["1234-5678", "9101-1121"],
            [
                {
                    "_index": "impact_factors",
                    "_id": "NAmRrYkB3O3HeJL6-cS3",
                    "_score": None,
                    "_source": {
                        "Scopus Source ID": 21100826277,
                        "Title": "Frontiers in Cell and Developmental Biology",
                        "Citation Count": 51808,
                        "Scholarly Output": 8178,
                        "Percent Cited": 78,
                        "CiteScore": 1.0,
                        "SNIP": 0.995,
                        "SJR": 1.418,
                        "Scopus ASJC Code (Sub-subject Area)": 1309,
                        "Scopus Sub-Subject Area": "Developmental Biology",
                        "Percentile": 69,
                        "RANK": 24,
                        "Rank Out Of": 78,
                        "Main Publisher": "Frontiers Media S.A.",
                        "Type": "j",
                        "Open Access": "YES",
                        "Quartile": 2,
                        "Top 10% (CiteScore Percentile)": False,
                        "URL Scopus Source ID": (
                            "https://www.scopus.com/sourceid/21100826277"
                        ),
                        "Print ISSN": None,
                        "E-ISSN": "12345678",
                    },
                    "sort": [48471],
                },
                {
                    "_index": "impact_factors",
                    "_id": "NAmRrYkB3O3HeJL6-cS3",
                    "_score": None,
                    "_source": {
                        "Scopus Source ID": 21100826277,
                        "Title": "Frontiers in Cell and Developmental Biology",
                        "Citation Count": 51808,
                        "Scholarly Output": 8178,
                        "Percent Cited": 78,
                        "CiteScore": 2.0,
                        "SNIP": 0.995,
                        "SJR": 1.418,
                        "Scopus ASJC Code (Sub-subject Area)": 1309,
                        "Scopus Sub-Subject Area": "Developmental Biology",
                        "Percentile": 69,
                        "RANK": 24,
                        "Rank Out Of": 78,
                        "Main Publisher": "Frontiers Media S.A.",
                        "Type": "j",
                        "Open Access": "YES",
                        "Quartile": 2,
                        "Top 10% (CiteScore Percentile)": False,
                        "URL Scopus Source ID": (
                            "https://www.scopus.com/sourceid/21100826277"
                        ),
                        "Print ISSN": None,
                        "E-ISSN": "91011121",
                    },
                    "sort": [48471],
                },
            ],
            {"1234-5678": 1.0, "9101-1121": 2.0},
        ),
        (["9999-9999"], [], {"9999-9999": None}),
        ([None], [], {None: None}),
    ],
)
@pytest.mark.asyncio
async def test_get_impact_factor(issns, return_value, response):
    """Test retrieving the impact factor."""
    fake_ds_client_instance = MagicMock(spec=AsyncBaseSearch)
    fake_ds_client_instance.get_available_indexes.return_value = [
        "something",
    ]
    fake_ds_client_instance.iter_document.return_value = AsyncIterator(return_value)

    impact_factors = await get_impact_factors(
        ds_client=fake_ds_client_instance,
        issns=issns,
        db_index_impact_factor="something",
    )

    assert impact_factors == response


@pytest.mark.parametrize(
    "issns,return_value,response",
    [
        (
            ["1234-5678", "9101-1121"],
            [
                {
                    "_index": "impact_factors",
                    "_id": "NAmRrYkB3O3HeJL6-cS3",
                    "_score": None,
                    "_source": {
                        "Scopus Source ID": 21100826277,
                        "Title": "Frontiers in Cell and Developmental Biology",
                        "Citation Count": 51808,
                        "Scholarly Output": 8178,
                        "Percent Cited": 78,
                        "CiteScore": 1.0,
                        "SNIP": 0.995,
                        "SJR": 1.418,
                        "Scopus ASJC Code (Sub-subject Area)": 1309,
                        "Scopus Sub-Subject Area": "Developmental Biology",
                        "Percentile": 69,
                        "RANK": 24,
                        "Rank Out Of": 78,
                        "Main Publisher": "Frontiers Media S.A.",
                        "Type": "j",
                        "Open Access": "YES",
                        "Quartile": 2,
                        "Top 10% (CiteScore Percentile)": False,
                        "URL Scopus Source ID": (
                            "https://www.scopus.com/sourceid/21100826277"
                        ),
                        "Print ISSN": None,
                        "E-ISSN": "12345678",
                    },
                    "sort": [48471],
                },
                {
                    "_index": "impact_factors",
                    "_id": "NAmRrYkB3O3HeJL6-cS3",
                    "_score": None,
                    "_source": {
                        "Scopus Source ID": 21100826277,
                        "Title": "Frontiers in Cell and Developmental Biology",
                        "Citation Count": 51808,
                        "Scholarly Output": 8178,
                        "Percent Cited": 78,
                        "CiteScore": 2.0,
                        "SNIP": 0.995,
                        "SJR": 1.418,
                        "Scopus ASJC Code (Sub-subject Area)": 1309,
                        "Scopus Sub-Subject Area": "Developmental Biology",
                        "Percentile": 69,
                        "RANK": 24,
                        "Rank Out Of": 78,
                        "Main Publisher": "Frontiers Media S.A.",
                        "Type": "j",
                        "Open Access": "YES",
                        "Quartile": 2,
                        "Top 10% (CiteScore Percentile)": False,
                        "URL Scopus Source ID": (
                            "https://www.scopus.com/sourceid/21100826277"
                        ),
                        "Print ISSN": None,
                        "E-ISSN": "91011121",
                    },
                    "sort": [48471],
                },
            ],
            {"1234-5678": 1.0, "9101-1121": 2.0},
        ),
        (["9999-9999"], [], {"9999-9999": None}),
        ([None], [], {None: None}),
    ],
)
@pytest.mark.asyncio
async def test_metadata_retriever_get_impact_factor(issns, return_value, response):
    """Test retrieving the impact factor."""
    fake_ds_client_instance = MagicMock(spec=AsyncBaseSearch)
    fake_ds_client_instance.get_available_indexes.return_value = [
        "something",
    ]
    fake_ds_client_instance.iter_document.return_value = AsyncIterator(return_value)

    retriever = MetaDataRetriever()
    retriever.schedule_bulk_request(
        get_impact_factors,
        **{
            "ds_client": fake_ds_client_instance,
            "issns": issns,
            "db_index_impact_factor": "something",
        },
    )
    impact_factors = await retriever.fetch()

    assert impact_factors["get_impact_factors"] == response


@pytest.mark.asyncio
async def test_get_citation_count(httpx_mock):
    """Test get_citation_count."""
    doi = "1"
    response = {"citationCount": 2}

    httpx_mock.add_response(
        url=(
            f"https://api.semanticscholar.org/graph/v1/paper/{doi}?fields=citationCount"
        ),
        method="GET",
        json=response,
    )
    async with httpx.AsyncClient() as client:
        result = await get_citation_count(doi, client)

    assert result == 2


@pytest.mark.asyncio
async def test_get_citation_count_no_doi():
    """Test get_citation_count."""
    doi = None

    async with httpx.AsyncClient() as client:
        result = await get_citation_count(doi, client)

    assert result is None


@pytest.mark.asyncio
async def test_metadata_retriever_get_citation_count(httpx_mock):
    """Test get_citation_counts."""
    dois = ["1", "2", "3"]
    responses = [{"citationCount": 2}, {"citationCount": 5}, {}]
    retriever = MetaDataRetriever()
    for doi, response in zip(dois, responses):
        httpx_mock.add_response(
            url=f"https://api.semanticscholar.org/graph/v1/paper/{doi}?fields=citationCount",
            method="GET",
            json=response,
        )
    async with httpx.AsyncClient() as client:
        retriever.schedule_non_bulk_requests(
            get_citation_count, dois, **{"httpx_client": client}
        )
        result = await retriever.fetch()

    assert result["get_citation_count"] == {"1": 2, "2": 5, "3": None}


@pytest.mark.asyncio
async def test_get_citation_counts_exception(httpx_mock):
    """Test get_citation_counts in case of failure."""
    # Test case with DOI that triggers httpx.HTTPError
    doi = "error_doi"

    httpx_mock.add_exception(
        url=(
            f"https://api.semanticscholar.org/graph/v1/paper/{doi}?fields=citationCount"
        ),
        exception=httpx.HTTPError(
            "DOI not found"
        ),  # Use httpx.HTTPError to trigger exception
    )
    async with httpx.AsyncClient() as client:
        result = await get_citation_count(doi, client)

    assert result is None


@pytest.mark.asyncio
async def test_reconstruct_abstract():
    fake_paragraph_1 = {
        "_index": "fake_index",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 2,
            "section": "Abstract",
            "text": "but the third one is even better.",
        },
    }
    fake_paragraph_2 = {
        "_index": "fake_index",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 0,
            "section": "Abstract",
            "text": (
                "This first paragraph contains the great first part of the great"
                " abstract, "
            ),
        },
    }
    fake_paragraph_3 = {
        "_index": "fake_index",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 1,
            "section": "Abstract",
            "text": "however i believe that this second part is amazing ",
        },
    }
    fake_ds_instance = MagicMock(spec=AsyncBaseSearch)

    fake_ds_instance.iter_document.return_value = AsyncIterator(
        [fake_paragraph_1, fake_paragraph_2, fake_paragraph_3]
    )

    abstract = await recreate_abstract("1234abcd", fake_ds_instance, "fake_index")

    assert (
        abstract
        == "This first paragraph contains the great first part of the great abstract,"
        " however i believe that this second part is amazing but the third one is"
        " even better."
    )


@pytest.mark.asyncio
async def test_recreate_abstract_with_db(get_testing_async_ds_client):
    ds_client, parameters = get_testing_async_ds_client

    index_doc = "test_paragraphs"

    await ds_client.create_index(
        index_doc, settings=parameters[-1], mappings=parameters[0]
    )

    fake_paragraph_1 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 2,
            "section": "Abstract",
            "text": "but the third one is even better.",
        },
    }
    fake_paragraph_2 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 0,
            "section": "Abstract",
            "text": (
                "This first paragraph contains the great first part of the great"
                " abstract, "
            ),
        },
    }
    fake_paragraph_3 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 1,
            "section": "Abstract",
            "text": "however i believe that this second part is amazing ",
        },
    }
    doc_bulk = [fake_paragraph_1, fake_paragraph_2, fake_paragraph_3]

    await ds_client.bulk(doc_bulk)

    await ds_client.client.indices.refresh()

    abstract = await recreate_abstract("1234abcd", ds_client, "test_paragraphs")

    assert (
        abstract
        == "This first paragraph contains the great first part of the great abstract,"
        " however i believe that this second part is amazing but the third one is"
        " even better."
    )


@pytest.mark.asyncio
async def test_metadata_retriever_recreate_abstract_with_db(
    get_testing_async_ds_client,
):
    ds_client, parameters = get_testing_async_ds_client

    index_doc = "test_paragraphs"

    await ds_client.create_index(
        index_doc, settings=parameters[-1], mappings=parameters[0]
    )

    fake_paragraph_1 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 2,
            "section": "Abstract",
            "text": "but the third one is even better.",
        },
    }
    fake_paragraph_2 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 0,
            "section": "Abstract",
            "text": (
                "This first paragraph contains the great first part of the great"
                " abstract, "
            ),
        },
    }
    fake_paragraph_3 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "1234abcd",
            "paragraph_id": 1,
            "section": "Abstract",
            "text": "however i believe that this second part is amazing ",
        },
    }
    fake_paragraph_4 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "article2",
            "paragraph_id": 2,
            "section": "Abstract",
            "text": " it contains a bad abstract.",
        },
    }
    fake_paragraph_5 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "article2",
            "paragraph_id": 0,
            "section": "Abstract",
            "text": "This is a new article,",
        },
    }
    fake_paragraph_6 = {
        "_index": "test_paragraphs",
        "_source": {
            "article_id": "article2",
            "paragraph_id": 1,
            "section": "Abstract",
            "text": " it is quite a good one but",
        },
    }
    doc_bulk = [fake_paragraph_1, fake_paragraph_2, fake_paragraph_3]

    await ds_client.bulk(doc_bulk)

    doc_bulk = [fake_paragraph_4, fake_paragraph_5, fake_paragraph_6]

    await ds_client.bulk(doc_bulk)

    await ds_client.client.indices.refresh()

    retriever = MetaDataRetriever()
    retriever.schedule_non_bulk_requests(
        recreate_abstract,
        ["1234abcd", "article2"],
        **{"ds_client": ds_client, "db_index_paragraphs": "test_paragraphs"},
    )

    abstract = await retriever.fetch()

    assert (
        abstract["recreate_abstract"]["1234abcd"]
        == "This first paragraph contains the great first part of the great abstract,"
        " however i believe that this second part is amazing but the third one is"
        " even better."
    )

    assert (
        abstract["recreate_abstract"]["article2"]
        == "This is a new article, it is quite a good one but it contains a bad"
        " abstract."
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ["issn", "name"], [("1234-5678", "great_journal"), ("3426-1936", "bad_journal")]
)
async def test_get_journal_name(httpx_mock, issn, name):
    # If ISSN is None
    async with httpx.AsyncClient() as client:
        name_found = await get_journal_name(None, client)
    assert name_found is None

    response = {
        "@context": "http://schema.org/",
        "@id": f"https://portal.issn.org/resource/ISSN/{issn}",
        "@type": "Periodical",
        "exampleOfWork": {
            "@id": f"https://portal.issn.org/resource/ISSN-L/{issn}",
            "@type": "CreativeWork",
            "workExample": [
                {"@id": f"https://portal.issn.org/resource/ISSN/{issn}", "name": name}
            ],
        },
        "issn": issn,
        "identifier": [
            {
                "@type": "PropertyValue",
                "name": "ISSN",
                "value": issn,
                "description": "Valid",
            },
            {
                "@type": "PropertyValue",
                "name": "ISSN-L",
                "value": issn,
                "description": "Valid",
            },
        ],
        "name": name,
        "alternateName": "eNeurologicalSci.",
        "publication": {
            "@id": f"https://portal.issn.org/resource/ISSN/{issn}#ReferencePublicationEvent",
            "@type": "PublicationEvent",
            "location": {
                "@id": "https://www.iso.org/obp/ui/#iso:code:3166:NL",
                "@type": "Country",
                "name": "Netherlands",
            },
        },
    }
    response = (
        '<script type="application/ld+json">' + json.dumps(response) + "</script>"
    )
    httpx_mock.add_response(
        url=f"https://portal.issn.org/resource/ISSN/{issn}",
        method="GET",
        text=response,
    )
    async with httpx.AsyncClient() as client:
        name_found = await get_journal_name(issn, client)
    assert name == name_found

    httpx_mock.reset(assert_all_responses_were_requested=False)

    httpx_mock.add_response(
        url=f"https://portal.issn.org/resource/ISSN/{issn}",
        method="GET",
        text=response,
        status_code=404,
    )
    async with httpx.AsyncClient() as client:
        name = await get_journal_name(issn, client)
    assert not name

    httpx_mock.reset(assert_all_responses_were_requested=False)

    httpx_mock.add_response(
        url=f"https://portal.issn.org/resource/ISSN/{issn}",
        method="GET",
        text="terrible response.",
        status_code=200,
    )
    async with httpx.AsyncClient() as client:
        name = await get_journal_name(issn, client)
    assert not name


@pytest.mark.asyncio
async def test_journal_name_timeout(httpx_mock):
    httpx_mock.add_exception(httpx.PoolTimeout("Timeout"))
    async with httpx.AsyncClient() as client:
        name = await get_journal_name("1234-5678", client)
    assert name is None


@pytest.mark.asyncio
async def test_journal_name_http_exception(httpx_mock):
    httpx_mock.add_exception(httpx.HTTPError("This is the error"))
    async with httpx.AsyncClient() as client:
        name = await get_journal_name("1234-5678", client)
    assert name is None


@pytest.mark.asyncio
async def test_metadata_retriever_get_journal_name(httpx_mock):
    issns = ["1234-5678", "3426-1936"]
    names = ["great_journal", "bad_journal"]
    responses = [
        {
            "@context": "http://schema.org/",
            "@id": f"https://portal.issn.org/resource/ISSN/{issn}",
            "@type": "Periodical",
            "exampleOfWork": {
                "@id": f"https://portal.issn.org/resource/ISSN-L/{issn}",
                "@type": "CreativeWork",
                "workExample": [
                    {
                        "@id": f"https://portal.issn.org/resource/ISSN/{issn}",
                        "name": name,
                    }
                ],
            },
            "issn": issn,
            "identifier": [
                {
                    "@type": "PropertyValue",
                    "name": "ISSN",
                    "value": issn,
                    "description": "Valid",
                },
                {
                    "@type": "PropertyValue",
                    "name": "ISSN-L",
                    "value": issn,
                    "description": "Valid",
                },
            ],
            "name": name,
            "alternateName": "eNeurologicalSci.",
            "publication": {
                "@id": f"https://portal.issn.org/resource/ISSN/{issn}#ReferencePublicationEvent",
                "@type": "PublicationEvent",
                "location": {
                    "@id": "https://www.iso.org/obp/ui/#iso:code:3166:NL",
                    "@type": "Country",
                    "name": "Netherlands",
                },
            },
        }
        for (issn, name) in zip(issns, names)
    ]
    for issn, response in zip(issns, responses):
        response = (
            '<script type="application/ld+json">' + json.dumps(response) + "</script>"
        )
        httpx_mock.add_response(
            url=f"https://portal.issn.org/resource/ISSN/{issn}",
            method="GET",
            text=response,
        )
    retriever = MetaDataRetriever()
    async with httpx.AsyncClient() as client:
        retriever.schedule_non_bulk_requests(
            get_journal_name, issns, **{"httpx_client": client}
        )
        name_found = await retriever.fetch()
    assert name_found["get_journal_name"] == dict(zip(issns, names))


@pytest.mark.asyncio
async def test_metadata_retriever_external_apis(get_testing_async_ds_client):
    ds_client, _ = get_testing_async_ds_client
    contexts = [
        {
            "journal": "issn1",
            "doi": "doi1",
            "article_id": "fake",
        }
    ]

    async def get_journal_name(issn, httpx_client):
        return "journalname1"

    async def get_citation_count(doi, httpx_client):
        return "citationcount1"

    with (
        mock.patch(
            "scholarag.retrieve_metadata.get_impact_factors"  # not interesting for this test
        ),
        mock.patch(
            "scholarag.retrieve_metadata.recreate_abstract"  # not interesting for this test
        ),
        mock.patch(
            "scholarag.retrieve_metadata.get_citation_count",
            new=get_citation_count,
        ),
        mock.patch(
            "scholarag.retrieve_metadata.get_journal_name",
            new=get_journal_name,
        ),
    ):
        retriever = MetaDataRetriever(external_apis=False)
        async with httpx.AsyncClient() as client:
            fetched_metadata = await retriever.retrieve_metadata(
                contexts, ds_client, "impact_factors", "paragraphs", client
            )
        assert "get_journal_name" not in fetched_metadata
        assert "get_citation_count" not in fetched_metadata

        retriever = MetaDataRetriever(external_apis=True)
        async with httpx.AsyncClient() as client:
            fetched_metadata = await retriever.retrieve_metadata(
                contexts, ds_client, "impact_factors", "paragraphs", client
            )
        assert "get_journal_name" in fetched_metadata
        assert "get_citation_count" in fetched_metadata



"""Base classes for document stores."""

import logging
from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator, AsyncIterable, Iterable
from typing import Any

from pydantic import BaseModel, ConfigDict, SecretStr

logger = logging.getLogger(__name__)


class BaseSearch(BaseModel, ABC):
    """Base class for sync document stores."""

    host: str
    port: int
    user: str | None = None
    password: SecretStr | None = None
    use_ssl_and_verify_certs: bool | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    @abstractmethod
    def get_available_indexes(self) -> list[str]:
        """Return all available indexes."""

    @abstractmethod
    def remove_index(self, index: str) -> None:
        """Remove an index."""

    @abstractmethod
    def get_document(self, index: str, doc_id: str) -> dict[str, Any]:
        """Return a document.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_id
            ID under which the document is indexed.

        Returns
        -------
        doc
            Document retrieved.
        """

    @abstractmethod
    def get_documents(self, index: str, doc_ids: list[str]) -> list[dict[str, Any]]:
        """Return a list of documents.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_ids
            list of ids under which the documents are indexed.

        Returns
        -------
        docs
            list of document retrieved.
        """

    @staticmethod
    @abstractmethod
    def _process_search_hits(res: Any) -> list[dict[str, Any]]:
        """Process search hits.

        Parameters
        ----------
        res
            Result of a DB query.

        Returns
        -------
        out
            Processed results.
        """

    @abstractmethod
    def get_index_mappings(self, index: str) -> dict[str, Any]:
        """Return an index mapping."""

    @abstractmethod
    def create_index(
        self,
        index: str,
        settings: dict[str, Any] | None,
        mappings: dict[str, Any] | None,
    ) -> None:
        """Create a new index."""

    @abstractmethod
    def add_fields(
        self,
        index: str,
        settings: dict[str, Any] | None = None,
        mapping: dict[str, Any] | None = None,
    ) -> None:
        """Update the index with a new mapping and settings."""

    @abstractmethod
    def count_documents(self, index: str, query: dict[str, Any] | None = None) -> int:
        """Return the number of documents in an index."""

    @abstractmethod
    def exists(self, index: str, doc_id: str) -> bool:
        """Return True if this document exists in the index."""

    @abstractmethod
    def iter_document(
        self, index: str, query: dict[str, Any] | None = None, size: int = 1000
    ) -> Iterable[dict[str, Any]]:
        """Scan the documents matching a specific query."""

    @abstractmethod
    def add_document(
        self, index: str, doc: dict[str, Any], doc_id: str | None = None
    ) -> None:
        """Index a document."""

    @abstractmethod
    def bulk(self, actions: list[dict[str, Any]], **kwargs: Any) -> None:
        """Bulk upload of documents."""

    @abstractmethod
    def search(
        self,
        index: str,
        query: dict[str, Any] | None = None,
        size: int = 10,
        aggs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Wrap around the search api."""

    @abstractmethod
    def bm25_search(
        self,
        index_doc: str,
        query: str,
        filter_query: dict[str, Any] | None = None,
        k: int = 10,
    ) -> list[dict[str, Any]]:
        """BM25 search."""


class AsyncBaseSearch(BaseModel, ABC):
    """Base class for async document stores."""

    host: str
    port: int
    user: str | None = None
    password: SecretStr | None = None
    use_ssl: bool | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    async def get_available_indexes(self) -> list[str]:
        """Return all available indexes."""

    @abstractmethod
    async def remove_index(self, index: str) -> None:
        """Remove an index."""

    @abstractmethod
    async def get_document(self, index: str, doc_id: str) -> dict[str, Any]:
        """Return a document.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_id
            ID under which the document is indexed.

        Returns
        -------
        doc
            Document retrieved.
        """

    @abstractmethod
    async def get_documents(
        self, index: str, doc_ids: list[str]
    ) -> list[dict[str, Any]]:
        """Return a list of documents.

        Parameters
        ----------
        index
            DB index where documents are stored.
        doc_ids
            list of ids under which the documents are indexed.

        Returns
        -------
        docs
            list of document retrieved.
        """

    @staticmethod
    @abstractmethod
    def _process_search_hits(res: Any) -> list[dict[str, Any]]:
        """Process search hits.

        Parameters
        ----------
        res
            Result of a DB query.

        Returns
        -------
        out
            Processed results.
        """

    @abstractmethod
    async def get_index_mappings(self, index: str) -> dict[str, Any]:
        """Return an index mapping."""

    @abstractmethod
    async def create_index(
        self, index: str, settings: dict[str, Any], mappings: dict[str, Any]
    ) -> None:
        """Create a new index."""

    @abstractmethod
    async def add_fields(
        self,
        index: str,
        settings: dict[str, Any] | None = None,
        mapping: dict[str, Any] | None = None,
    ) -> None:
        """Update the index with a new mapping and settings."""

    @abstractmethod
    async def count_documents(
        self, index: str, query: dict[str, Any] | None = None
    ) -> int:
        """Return the number of documents in an index."""

    @abstractmethod
    async def exists(self, index: str, doc_id: str) -> bool:
        """Return True if this document exists in the index."""

    @abstractmethod
    def iter_document(
        self, index: str, query: dict[str, Any] | None = None, size: int = 1000
    ) -> AsyncIterable[dict[str, Any]] | AsyncGenerator[int, None]:
        """Scan the documents matching a specific query."""

    @abstractmethod
    async def add_document(
        self, index: str, doc: dict[str, Any], doc_id: str | None = None
    ) -> None:
        """Index a document."""

    @abstractmethod
    async def bulk(self, actions: list[dict[str, Any]], **kwargs: Any) -> None:
        """Bulk upload of documents."""

    @abstractmethod
    async def search(
        self,
        index: str,
        query: dict[str, Any] | None = None,
        size: int = 10,
        aggs: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Wrap around the search api."""

    @abstractmethod
    async def bm25_search(
        self,
        index_doc: str,
        query: str,
        filter_query: dict[str, Any] | None = None,
        k: int = 10,
    ) -> list[dict[str, Any]]:
        """BM25 search."""

    async def close(self) -> None:
        """Close the aiohttp session."""



"""Document store classes for Scholarag."""

from scholarag.document_stores.base import AsyncBaseSearch, BaseSearch
from scholarag.document_stores.elastic import AsyncElasticSearch, ElasticSearch
from scholarag.document_stores.open import AsyncOpenSearch, OpenSearch

__all__ = [
    "BaseSearch",
    "AsyncBaseSearch",
    "ElasticSearch",
    "AsyncElasticSearch",
    "OpenSearch",
    "AsyncOpenSearch",
]



"""Injectable dependencies for endpoints."""

import json
import logging
from enum import Enum
from functools import cache
from typing import Annotated, AsyncIterator

from elasticsearch_dsl import Q, Search
from fastapi import Depends, HTTPException, Query
from fastapi.security import OAuth2PasswordBearer
from httpx import AsyncClient, HTTPStatusError
from openai import AsyncOpenAI
from pydantic import constr
from starlette.status import HTTP_401_UNAUTHORIZED

from scholarag.app.config import Settings
from scholarag.document_stores import (
    AsyncBaseSearch,
    AsyncElasticSearch,
    AsyncOpenSearch,
)
from scholarag.generative_question_answering import GenerativeQAWithSources
from scholarag.services import CohereRerankingService, RetrievalService

logger = logging.getLogger(__name__)

journal_constraints = constr(pattern=r"^\d{4}-\d{3}[0-9X]$")

auth = OAuth2PasswordBearer(
    tokenUrl="/token",  # Will be overriden
    auto_error=False,
)


@cache
def get_settings() -> Settings:
    """Load all parameters.

    Note that this function is cached and environment will be read just once.
    """
    logger.info("Reading the environment and instantiating settings")
    return Settings()


async def get_httpx_client(
    settings: Annotated[Settings, Depends(get_settings)],
) -> AsyncIterator[AsyncClient]:
    """Manage the httpx client for the request."""
    client = AsyncClient(timeout=settings.metadata.timeout, verify=False)
    try:
        yield client
    finally:
        await client.aclose()


async def get_user_id(
    token: Annotated[str | None, Depends(auth)],
    settings: Annotated[Settings, Depends(get_settings)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
) -> str:
    """Validate JWT token and returns user ID."""
    if settings.keycloak.validate_token and settings.keycloak.user_info_endpoint:
        try:
            response = await httpx_client.get(
                settings.keycloak.user_info_endpoint,
                headers={"Authorization": f"Bearer {token}"},
            )
            response.raise_for_status()
            user_info = response.json()
            return user_info["sub"]
        except HTTPStatusError:
            raise HTTPException(
                status_code=HTTP_401_UNAUTHORIZED, detail="Invalid token."
            )
    else:
        return "dev"


def get_rts(settings: Annotated[Settings, Depends(get_settings)]) -> RetrievalService:
    """Get the retrieval service."""
    return RetrievalService(db_index_paragraphs=settings.db.index_paragraphs)


async def get_ds_client(
    settings: Annotated[Settings, Depends(get_settings)],
) -> AsyncIterator[AsyncBaseSearch]:
    """Get the Elasticsearch client."""
    try:
        if settings.db.password is None:
            password = None
        else:
            password = settings.db.password.get_secret_value()

        if settings.db.db_type == "elasticsearch":
            ds_client = AsyncElasticSearch(  # type: ignore
                host=settings.db.host,
                port=settings.db.port,
                user=settings.db.user,
                password=password,
            )
            yield ds_client
        else:
            ds_client = AsyncOpenSearch(  # type: ignore
                host=settings.db.host,
                port=settings.db.port,
                user=settings.db.user,
                password=password,
                use_ssl_and_verify_certs=True,
            )
            yield ds_client
    finally:
        await ds_client.close()


async def get_openai_client(
    settings: Annotated[Settings, Depends(get_settings)],
) -> AsyncIterator[AsyncOpenAI]:
    """Get async openai client for passthrough endpoint."""
    if settings.generative.openai.token:
        openai_client = AsyncOpenAI(
            api_key=settings.generative.openai.token.get_secret_value()
        )
    else:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.ENDPOINT_INACTIVE.value,
                "detail": "Couldn't connect to openai.",
            },
        )
    yield openai_client
    await openai_client.close()


def get_generative_qas(
    settings: Annotated[Settings, Depends(get_settings)],
    openai_client: Annotated[AsyncOpenAI, Depends(get_openai_client)],
) -> GenerativeQAWithSources:
    """Get the generative question answering service."""
    return GenerativeQAWithSources(
        client=openai_client,
        model=settings.generative.openai.model,
        temperature=settings.generative.openai.temperature,
        max_tokens=settings.generative.openai.max_tokens,
    )


async def get_reranker(
    settings: Annotated[Settings, Depends(get_settings)],
) -> AsyncIterator[CohereRerankingService | None]:
    """Get the reranking service."""
    if settings.reranking.cohere_token:
        reranker = CohereRerankingService(  # type: ignore
            api_key=settings.reranking.cohere_token.get_secret_value()
        )
    else:
        logger.warning(
            "The reranker is requested but the Cohere API key is not specified."
        )
        reranker = None
    yield reranker


def get_query_from_params(
    topics: Annotated[
        list[str] | None,
        Query(
            description="Keyword to be matched in text. AND matching (e.g. for TOPICS)."
        ),
    ] = None,
    regions: Annotated[
        list[str] | None,
        Query(
            description=(
                "Keyword to be matched in text. OR matching (e.g. for BRAIN_REGIONS)."
            )
        ),
    ] = None,
    article_types: Annotated[
        list[str] | None, Query(description="Article types allowed. OR matching")
    ] = None,
    authors: Annotated[
        list[str] | None, Query(description="List of authors. OR matching.")
    ] = None,
    journals: Annotated[  # type: ignore
        list[journal_constraints] | None,
        Query(
            description="List of journal issn. OR matching. Format: XXXX-XXXX",
        ),
    ] = None,
    date_from: Annotated[
        str | None,
        Query(
            description="Date lowerbound. Includes specified date. Format: YYYY-MM-DD",
            pattern=r"^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01])$",
        ),
    ] = None,
    date_to: Annotated[
        str | None,
        Query(
            description="Date upperbound. Includes specified date. Format: YYYY-MM-DD",
            pattern=r"^\d{4}-(0[1-9]|1[0-2])-(0[1-9]|[12]\d|3[01])$",
        ),
    ] = None,
) -> dict[str, str] | None:
    """Get the query parameters and generate an ES query for filtering."""
    search = Search()
    search_elems = []
    if topics:
        linked_tokens = [
            (
                " AND ".join(("(" + keyword + ")").split(" "))
                if len(keyword.split(" ")) >= 2
                else keyword
            )
            for keyword in topics
        ]
        topics_bool = f"({' AND '.join(linked_tokens)})"
        search_elems.append(topics_bool)

    if regions:
        linked_tokens = [
            (
                " AND ".join(("(" + keyword + ")").split(" "))
                if len(keyword.split(" ")) >= 2
                else keyword
            )
            for keyword in regions
        ]
        regions_bool = f"({' OR '.join(linked_tokens)})"
        search_elems.append(regions_bool)

    if topics or regions:
        q = Q(
            "query_string", default_field="text", query=f"{' AND '.join(search_elems)}"
        )
        search = search.query(q)
    if article_types:
        search = search.query(Q("terms", article_type=article_types))
    if authors:
        search = search.query(Q("terms", authors=authors))
    if journals:
        search = search.query(Q("terms", journal=journals))
    if date_from:
        search = search.query(Q("range", date={"gte": date_from}))
    if date_to:
        search = search.query(Q("range", date={"lte": date_to}))

    logger.info(
        f"Searching the database with the query {json.dumps(search.to_dict())}."
    )
    return None if not search.to_dict() else search.to_dict()["query"]


class ErrorCode(Enum):
    """ErrorCode class to be used when an error code has to be raised.

    Append new entries everytime a new error code is created.
    """

    NO_DB_ENTRIES = 1  # Has an HTTPS status code 500, since it is related to a problem in the backend.
    NO_ANSWER_FOUND = 2  # Has an HTTPS status code 500, since it is related to a problem in the backend.
    ENDPOINT_INACTIVE = 3  # Has an HTTPS status code 500, since it is related to a problem in the backend.
    INPUT_EXCEEDED_MAX_TOKENS = 4  # Has an HTTPS status code 413, since it happens when the inputs are too large.
    MAX_COHERE_REQUESTS = 5  # Has an HTTPS status code 500, since it is related to a problem in the backend.
    INCOMPLETE_ANSWER = 6  # Has an HTTPS status code 500, since it is related to a problem in the backend.
    ALREADY_EXISTS_IN_DB = 7



"""Endpoints for retrieving documents."""

import logging
import time
from typing import Annotated, Any

from fastapi import APIRouter, Depends, HTTPException, Query, Request
from fastapi_pagination import Page, paginate
from httpx import AsyncClient

from scholarag.app.config import Settings
from scholarag.app.dependencies import (
    ErrorCode,
    get_ds_client,
    get_httpx_client,
    get_query_from_params,
    get_reranker,
    get_rts,
    get_settings,
)
from scholarag.app.schemas import (
    ArticleCountResponse,
    ArticleMetadata,
    ParagraphMetadata,
    RetrievalRequest,
)
from scholarag.document_stores import AsyncBaseSearch
from scholarag.retrieve_metadata import MetaDataRetriever
from scholarag.services import CohereRerankingService, RetrievalService

router = APIRouter(prefix="/retrieval", tags=["Retrieval"])

logger = logging.getLogger(__name__)


@router.get(
    "/",
    response_model=list[ParagraphMetadata],
    responses={
        500: {
            "description": (
                f"If code={ErrorCode.NO_DB_ENTRIES.value}, no"
                " relevant context is found in the database."
            ),
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                },
            },
        }
    },
)
async def retrieval(
    rts: Annotated[RetrievalService, Depends(get_rts)],
    rs: Annotated[CohereRerankingService | None, Depends(get_reranker)],
    ds_client: Annotated[AsyncBaseSearch, Depends(get_ds_client)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
    filter_query: Annotated[dict[str, Any], Depends(get_query_from_params)],
    settings: Annotated[Settings, Depends(get_settings)],
    request: RetrievalRequest = Depends(),  # noqa: B008
) -> list[ParagraphMetadata]:
    """Answer the query with most relevant paragraphs.
    \f

    Parameters
    ----------
    request
        Body of the request sent to the server.
    rts
        Retrieval service.
    rs
        Reranker service.
    ds_client
        Document store client
    httpx_client
        HTTP Client.
    filter_query
        Filtering query for paragraph retrieval.
    settings
        Global settings of the application

    Returns
    -------
        A list of article titles and paragraphs.
    """  # noqa: D301, D400, D205
    start = time.time()
    logger.info("Finding documents ...")

    contexts = await rts.arun(
        ds_client=ds_client,
        query=request.query,
        retriever_k=request.retriever_k,
        db_filter=filter_query,
        max_length=settings.retrieval.max_length,
    )
    if not contexts:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.NO_DB_ENTRIES.value,
                "detail": (
                    "No document found. Modify the filters or the query and try again."
                ),
            },
        )
    logger.info(f"Semantic search took {time.time() - start}s")

    if rs is None or not request.use_reranker:
        indices = tuple(range(len(contexts)))
        scores = None
    else:
        contexts, _, scores, indices = await rs.rerank(
            query=request.query,
            contexts=contexts,
            reranker_k=request.reranker_k,
        )

    metadata_retriever = MetaDataRetriever(
        external_apis=settings.metadata.external_apis
    )

    fetched_metadata = await metadata_retriever.retrieve_metadata(
        contexts,
        ds_client,
        settings.db.index_journals,
        settings.db.index_paragraphs,
        httpx_client,
    )

    citedby_counts = fetched_metadata.get("get_citation_count", {})
    journal_names = fetched_metadata.get("get_journal_name", {})
    impact_factors = fetched_metadata["get_impact_factors"]
    abstracts = fetched_metadata["recreate_abstract"]

    responses = []
    for i, context in enumerate(contexts):
        journal_issn = context.get("journal")
        metadata = {
            "article_title": context["title"],
            "section": context["section"],
            "paragraph": context["text"],
            "journal_issn": context.get("journal"),
            "date": context.get("date"),
            "article_id": context["article_id"],
            "ds_document_id": str(context["document_id"]),
            "article_doi": context["doi"],
            "pubmed_id": context["pubmed_id"],
            "article_authors": context["authors"],
            "article_type": context["article_type"],
            "context_id": indices[i],
            "reranking_score": scores[i] if scores is not None else None,
            # on the fly metadata
            "abstract": abstracts[context.get("article_id")],
            "journal_name": journal_names.get(journal_issn),
            "impact_factor": impact_factors[journal_issn],
            "cited_by": citedby_counts.get(context.get("doi")),
        }

        responses.append(ParagraphMetadata(**metadata))
    return responses


@router.get("/article_count")
async def article_count(
    ds_client: Annotated[AsyncBaseSearch, Depends(get_ds_client)],
    filter_query: Annotated[dict[str, Any], Depends(get_query_from_params)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> ArticleCountResponse:
    """Article count based on keyword matching.
    \f

    Parameters
    ----------
    ds_client
        Document store client
    filter_query
        Filtering query for paragraph retrieval.
    settings
        Global settings of the application

    Returns
    -------
        A list of article titles and paragraphs.
    """  # noqa: D301, D400, D205
    start = time.time()
    logger.info("Finding unique articles matching the query ...")
    query = filter_query

    # Aggregation query.
    aggs = {
        "article_count": {
            "cardinality": {
                "field": "article_id",
                "precision_threshold": (
                    10000
                ),  # Might be approximative if more documents.
            },
        }
    }
    results = await ds_client.search(
        index=settings.db.index_paragraphs, query=query, size=0, aggs=aggs
    )
    article_count = results["aggregations"]["article_count"]["value"]
    logger.info(f"unique article retrieval took: {time.time() - start}s")
    return ArticleCountResponse(article_count=article_count)


@router.get(
    "/article_listing",
    response_model=Page[ArticleMetadata],
    responses={
        500: {
            "description": (
                "No article corresponding to the filters found. If"
                f" code={ErrorCode.NO_DB_ENTRIES.value}, no article matching filter"
                " found."
            ),
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                },
            },
        }
    },
)
async def article_listing(
    request: Request,
    ds_client: Annotated[AsyncBaseSearch, Depends(get_ds_client)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
    filter_query: Annotated[dict[str, Any], Depends(get_query_from_params)],
    settings: Annotated[Settings, Depends(get_settings)],
    number_results: Annotated[
        int | None,
        Query(description="Number of results to return. Max 10 000.", ge=1, le=10_000),
    ] = 100,
    sort_by_date: Annotated[
        bool | None,
        Query(
            description=(
                "Sort by date True or False. If the latter, we sort by relevance"
            )
        ),
    ] = False,
) -> Page[ArticleMetadata]:
    """Article id listing based on keyword matching.
    \f

    Parameters
    ----------
    request
        Request sent by the user.
    ds_client
        Document store client.
    httpx_client
        HTTP Client.
    filter_query
        Filtering query for paragraph retrieval.
    settings
        Global settings of the application
    number_results
        Number of results to return. Max 10 000.
    sort_by_date
        Sort by date True or False. If the latter, we sort by relevance

    Returns
    -------
        A list of article titles and paragraphs.
    """  # noqa: D301, D400, D205
    start = time.time()
    logger.info("Finding unique articles matching the query ...")

    params = request.query_params
    topics = params.getlist("topics")
    regions = params.getlist("regions")

    if topics or regions:
        query = {
            "query": {
                "bool": {
                    "must": [
                        {"match": {"text": " ".join(topics) + " " + " ".join(regions)}}
                    ],
                    "filter": filter_query,
                }
            },
        }
    elif filter_query and not topics and not regions:
        query = {"query": {"match_all": {}}, "filter": filter_query}
    else:
        query = {"query": {"match_all": {}}}

    aggs: dict[str, Any] = {
        "relevant_ids": {
            "terms": {
                "size": number_results,
                "field": "article_id",
                "order": {"score": "desc"},
            },
            "aggs": {
                "score": {"max": {"script": "_score"}},
                "ids_hit": {"top_hits": {"size": 1}},
            },
        }
    }

    if sort_by_date:
        aggs["relevant_ids"]["aggs"]["score"]["max"] = {"field": "date"}

    results = await ds_client.search(
        index=settings.db.index_paragraphs, query=query, size=0, aggs=aggs
    )
    logger.info(f"unique article retrieval took: {time.time() - start}s")

    docs = [
        res["ids_hit"]["hits"]["hits"][0]["_source"]
        for res in results["aggregations"]["relevant_ids"]["buckets"]
    ]

    if len(docs) == 0:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.NO_DB_ENTRIES.value,
                "detail": (
                    "No article found corresponding to the filters. You can use the"
                    " article_count endpoint to check which set of filters actually"
                    " returns documents. If the article_count endpoint returned a non"
                    " zero count with the same filters, report the issue to the"
                    " Machine Learning team on Slack."
                ),
            },
        )

    metadata_retriever = MetaDataRetriever(
        external_apis=settings.metadata.external_apis
    )

    start_2 = time.time()
    fetched_metadata = await metadata_retriever.retrieve_metadata(
        docs,
        ds_client,
        settings.db.index_journals,
        settings.db.index_paragraphs,
        httpx_client,
    )
    logger.info(
        f"Retrieved metadata of {len(docs)} articles in {time.time() - start_2}s."
    )
    citedby_counts = fetched_metadata.get("get_citation_count", {})
    journal_names = fetched_metadata.get("get_journal_name", {})
    impact_factors = fetched_metadata["get_impact_factors"]
    abstracts = fetched_metadata["recreate_abstract"]

    responses = []
    for context in docs:
        journal_issn = context.get("journal")
        metadata = {
            "article_title": context["title"],
            "journal_issn": context.get("journal"),
            "date": context.get("date"),
            "article_id": context["article_id"],
            "article_doi": context["doi"],
            "pubmed_id": context["pubmed_id"],
            "article_authors": context["authors"],
            "article_type": context["article_type"],
            # on the fly metadata
            "abstract": abstracts[context.get("article_id")],
            "journal_name": journal_names.get(journal_issn),
            "impact_factor": impact_factors[journal_issn],
            "cited_by": citedby_counts.get(context.get("doi")),
        }

        responses.append(ArticleMetadata(**metadata))
    return paginate(responses)



"""Impact factor retrieval."""

from __future__ import annotations

import asyncio
import json
import logging
import ssl
from typing import Any, Callable, Coroutine

import httpx

from scholarag.document_stores import AsyncBaseSearch

logger = logging.getLogger(__name__)


class MetaDataRetriever:
    """Class to send API calls async."""

    def __init__(
        self, tasks: list[asyncio.Task[Any]] | None = None, external_apis: bool = True
    ) -> None:
        if tasks is None:
            self.tasks = []
        else:
            self.tasks = tasks
        self.external_apis = external_apis

    def schedule_non_bulk_requests(
        self,
        coro: Callable[..., Coroutine[Any, Any, str | int | None]],
        data_list: list[str],
        **kwargs: Any,
    ) -> None:
        """Schedule multiple individual API calls."""
        # Remove duplicates before sending the requests.
        data_list = list(set(data_list))

        tasks = [asyncio.create_task(coro(data, **kwargs)) for data in data_list]
        self.tasks.append(
            asyncio.create_task(
                self.execute_sub_tasks(tasks, data_list), name=coro.__name__
            )
        )

    def schedule_bulk_request(
        self,
        coro: Callable[
            ...,
            Coroutine[
                Any, Any, list[str | None] | dict[str, Any] | list[dict[str, Any]]
            ],
        ],
        **kwargs: Any,
    ) -> None:
        """Schedule one API call that is compatible with bulk requests."""
        self.tasks.append(asyncio.create_task(coro(**kwargs), name=coro.__name__))

    @staticmethod
    async def execute_sub_tasks(
        tasks: list[asyncio.Task[Any]], data_list: list[str]
    ) -> dict[str, Any]:
        """Execute sub-tasks."""
        result = await asyncio.gather(*tasks)
        result_dict = dict(zip(data_list, result))
        return result_dict

    async def fetch(
        self,
    ) -> dict[str, Any]:
        """Run the API calls async."""
        task_names = [task.get_name() for task in self.tasks]
        results = await asyncio.gather(*self.tasks)
        return dict(zip(task_names, results))

    async def retrieve_metadata(
        self,
        contexts: list[dict[str, Any]],
        ds_client: AsyncBaseSearch,
        db_if_articles: str | None,
        db_index_paragraphs: str,
        httpx_client: httpx.AsyncClient,
    ) -> dict[str, Any]:
        """Schedule the metadata retrieval and execute all of the tasks."""
        issns = [
            context["journal"] if "journal" in context else None for context in contexts
        ]
        dois = [context["doi"] for context in contexts if context["doi"] is not None]
        article_ids = [context["article_id"] for context in contexts]

        self.schedule_bulk_request(
            get_impact_factors,
            **{
                "ds_client": ds_client,
                "issns": issns,
                "db_index_impact_factor": db_if_articles,
            },
        )
        self.schedule_non_bulk_requests(
            recreate_abstract,
            article_ids,
            **{
                "ds_client": ds_client,
                "db_index_paragraphs": db_index_paragraphs,
            },
        )

        if self.external_apis:
            self.schedule_non_bulk_requests(
                get_citation_count, dois, **{"httpx_client": httpx_client}
            )
            self.schedule_non_bulk_requests(
                get_journal_name, issns, **{"httpx_client": httpx_client}
            )

        fetched_metadata = await self.fetch()
        return fetched_metadata


async def get_citation_count(
    doi: str | None, httpx_client: httpx.AsyncClient
) -> int | None:
    """Get the citation count for a given DOI.

    Parameters
    ----------
    doi : str | None
        The DOI of the paper.
    httpx_client : httpx.AsyncClient
        Async HTTPX Client.

    Returns
    -------
    int or None
        The number of citations for the paper if found, else None.
    """
    if doi is None:
        return None

    logger.info(f"Sending citation count request with doi {doi}.")
    ss_base_url = "https://api.semanticscholar.org/graph/v1/paper"
    ss_params = {"fields": "citationCount"}

    url = f"{ss_base_url}/{doi}"

    try:
        response = await httpx_client.get(url, params=ss_params)
        logger.debug(f"Obtained response for doi {doi}.")
        response.raise_for_status()
        citation_data = response.json()
        return citation_data.get("citationCount", None)

    except httpx.HTTPError as e:
        if hasattr(e, "detail") and hasattr(e.detail, "response"):
            logger.error(
                "Received non-success HTTP status code:"
                f" {e.detail.response.status_code}"
            )
            logger.error(f"Response content: {e.detail.response.text}")
        else:
            logger.info(f"HTTPError occurred, doi: {doi}")
        return None


async def get_impact_factors(
    ds_client: AsyncBaseSearch,
    issns: list[str],
    db_index_impact_factor: str | None,
) -> dict[str, float | None]:
    """Run impact factor search asynchronously.

    Parameters
    ----------
    ds_client
        Client to connect to Elasticsearch to get document data.
    issns
        Unique identifiers for journals.
    db_index_impact_factor
        Elasticsearch index for paragraphs

    Returns
    -------
    issn_to_citescore : dict[str, float | None]
        Dictionary containing the ISSN as keys and citescore as values.
    """
    logger.info("Running impact factor search")

    available_indexes = await ds_client.get_available_indexes()
    if (
        db_index_impact_factor is None
        or db_index_impact_factor not in available_indexes
    ):
        logger.warning(
            f"Skipping impact factor retrieval. The index {db_index_impact_factor} does"
            " not exist."
        )
        return {issn: None for issn in issns}

    filtered_issns = [issn[:4] + issn[5:] for issn in issns if issn is not None]
    query = {"query": {"terms": {"E-ISSN": filtered_issns}}}
    results = ds_client.iter_document(db_index_impact_factor, query)
    issn_to_citescore: dict[str, float | None] = {}

    for issn in issns:
        issn_to_citescore[issn] = None

    async for result in results:
        source = result["_source"]  # type: ignore
        issn = source["E-ISSN"][:4] + "-" + source["E-ISSN"][4:]
        issn_to_citescore[issn] = source["CiteScore"]
    logger.debug("Done retrieving impact factors.")
    return issn_to_citescore


async def get_journal_name(
    issn: str | None, httpx_client: httpx.AsyncClient
) -> str | None:
    """Fetch the journal name based on its issn."""
    if issn is None:
        return None
    logger.info(f"Fetching journal name based on issn: {issn}")
    url = f"https://portal.issn.org/resource/ISSN/{issn}"

    try:
        response = await httpx_client.get(url, follow_redirects=True)
        response.raise_for_status()
    except (httpx.ReadTimeout, httpx.PoolTimeout):
        logger.warning(f"Request to {url} timed out.")
        return None
    except ssl.SSLCertVerificationError:
        logger.error(f"SSL certificate issue with url {url}")
        return None
    except httpx.HTTPError:
        logger.warning(f"Request to {url} failed.")
        return None

    start = response.text.find('<script type="application/ld+json">')
    end = response.text.find("</script>", start)
    curated_text = (
        response.text[start:end]
        .replace('<script type="application/ld+json">', "")
        .replace("</script>", "")
    )
    try:
        response_dict = json.loads(curated_text)
    except json.decoder.JSONDecodeError:
        logger.warning("Could not load the API response as a JSON.")
        return None
    name = response_dict.get("name")
    logger.debug(f"Done retrieving journal name based on ISSN {issn}.")

    return name


async def recreate_abstract(
    article_id: str, ds_client: AsyncBaseSearch, db_index_paragraphs: str
) -> str | None:
    """Recreate the abstract from the article_id of the context."""
    logger.info(f"Reconstructing abstract of article {article_id}.")
    query = {
        "query": {
            "bool": {
                "must": [
                    {"term": {"article_id": article_id}},
                    {"term": {"section": "Abstract"}},
                ]
            }
        }
    }
    paragraphs = ds_client.iter_document(db_index_paragraphs, query)
    sorted_paragraphs = sorted(
        [paragraph["_source"] async for paragraph in paragraphs],  # type: ignore
        key=lambda x: x["paragraph_id"],
    )
    sorted_texts = [paragraph["text"] for paragraph in sorted_paragraphs]
    if len(sorted_texts) == 0:
        return None
    abstract = "".join(sorted_texts)
    logger.debug(f"Done reconstructing abstract of article {article_id}.")
    return abstract



"""Test journal_suggestion related functions."""

import pytest
from scholarag.utils import find_files, format_issn


@pytest.mark.parametrize(
    "source_issns,formatted_issns",
    [
        (None, None),
        ("12345678", "1234-5678"),
        ("345678", "0034-5678"),
        ("12345678 91011121", "1234-5678 9101-1121"),
    ],
)
def test_format_issn(source_issns, formatted_issns):
    """Test issn formatting."""
    result_issns = format_issn(source_issns)
    assert result_issns == formatted_issns


def test_format_issn_error():
    """Test issn formatting raising issue."""
    with pytest.raises(ValueError):
        _ = format_issn("wrong-issn")


def test_find_files(tmp_path):
    """Test filtering files."""
    # Create a fake file
    file1_path = tmp_path / "file1.json"
    file1_path.touch()
    file2_path = tmp_path / "file2.txt"
    file2_path.touch()
    file3_path = tmp_path / "dir" / "file3.json"
    file3_path.parent.mkdir()
    file3_path.touch()
    file4_path = tmp_path / "dir" / "file4.txt"
    file4_path.touch()

    files = find_files(file1_path, False)
    assert len(files) == 1

    files = find_files(tmp_path, False, r".*\.json$")
    assert len(files) == 1

    files = find_files(tmp_path, False)
    assert len(files) == 2

    files = find_files(tmp_path, True, r".*\.json$")
    assert len(files) == 2


def test_find_files_errors(tmp_path):
    """Test filtering files."""
    with pytest.raises(ValueError):
        _ = find_files(tmp_path, False, "")

    # Wrong file mention
    file_path = tmp_path / "doesnotexist.json"
    with pytest.raises(ValueError):
        _ = find_files(file_path, False)



"""Endpoints response and request schemas."""

from pydantic import BaseModel, Field


class JournalSuggestionRequest(BaseModel):
    """Request for the journal suggestion endpoint."""

    keywords: str
    limit: int = Field(100, ge=1, le=10000)


class JournalSuggestionResponse(BaseModel):
    """Class representing Journal metadata."""

    title: str
    citescore: float | None
    eissn: str | None
    snip: float | None
    sjr: float | None
    print_issn: str | None


class ArticleTypeSuggestionResponse(BaseModel):
    """Class representing article types in the db."""

    article_type: str
    docs_in_db: int


class AuthorSuggestionResponse(BaseModel):
    """Class representing authors."""

    name: str


class AuthorSuggestionRequest(BaseModel):
    """Request for the author suggestion endpoint."""

    name: str
    limit: int = Field(100, ge=1, le=1000)


class PassthroughRequest(BaseModel):
    """Request to the passthrough endpoint."""

    query: str


class RetrievalRequest(PassthroughRequest, extra="forbid"):
    """Request for the raw retrieval endpoint."""

    retriever_k: int = Field(700, ge=1, le=2000)
    use_reranker: bool = True
    reranker_k: int = Field(8, ge=1, le=100)


class GenerativeQARequest(RetrievalRequest, extra="forbid"):
    """Request for the QA endpoint."""


class ArticleCountResponse(BaseModel, extra="forbid"):
    """Response of the article_count endpoint."""

    article_count: int


class ArticleMetadata(BaseModel, extra="ignore"):
    """Metadata for an article."""

    article_title: str
    article_authors: list[str]
    article_id: str

    # Possibly missing
    article_doi: str | None = None
    pubmed_id: str | None = None
    date: str | None = None
    article_type: str | None = None
    journal_issn: str | None = None

    # Added dynamically
    journal_name: str | None = None
    cited_by: int | None = None
    impact_factor: float | None = None
    abstract: str | None = None


class DocumentMetadata(ArticleMetadata, extra="ignore"):
    """Metadata for a document retrieved directly from the db."""

    paragraph: str
    ds_document_id: str

    # Possibly missing
    section: str | None = None


class ParagraphMetadata(DocumentMetadata, extra="ignore"):
    """Metadata for a paragraphs retrieved through the retrieval service."""

    context_id: int  # Position of the context after retrieval (from db).

    # Possibly missing
    reranking_score: float | None = None


class PassthroughResponse(BaseModel):
    """Response for the passthrough endpoint."""

    answer: str


class GenerativeQAResponse(BaseModel):
    """Response for the generative QA endpoint."""

    answer: str | None
    raw_answer: str
    metadata: list[ParagraphMetadata]



"""Endpoints for suggestions."""

import logging
import re
import time
from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException

from scholarag.app.config import Settings
from scholarag.app.dependencies import ErrorCode, get_ds_client, get_settings
from scholarag.app.schemas import (
    ArticleTypeSuggestionResponse,
    AuthorSuggestionRequest,
    AuthorSuggestionResponse,
    JournalSuggestionRequest,
    JournalSuggestionResponse,
)
from scholarag.document_stores import AsyncBaseSearch
from scholarag.utils import format_issn

router = APIRouter(prefix="/suggestions", tags=["Suggestions"])

logger = logging.getLogger(__name__)


@router.get(
    "/journal",
    response_model=list[JournalSuggestionResponse],
    responses={
        500: {
            "description": f"code={ErrorCode.NO_DB_ENTRIES.value} No journal found.",
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                },
            },
        }
    },
)
async def journal_suggestion(
    ds_client: Annotated[AsyncBaseSearch, Depends(get_ds_client)],
    settings: Annotated[Settings, Depends(get_settings)],
    request: JournalSuggestionRequest = Depends(),  # noqa: B008
) -> list[JournalSuggestionResponse]:
    """Suggest journals with metadata based on keywords.
    \f

    Parameters
    ----------
    request
        Body of the request sent to the server.
    ds_client
        Document store client
    settings
        Global settings of the application

    Returns
    -------
        A list of answers with journal metadata added.
    """  # noqa: D301, D400, D205
    if settings.db.index_journals is None:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.ENDPOINT_INACTIVE.value,
                "detail": (
                    "This endpoint is inactive. Please provide the "
                    "SCHOLARAG__DB__INDEX_JOURNALS environment variable."
                ),
            },
        )
    start = time.time()
    logger.info("Finding journals ...")

    keywords = request.keywords.split()

    match_queries = [
        {"regexp": {"Title": {"value": f".*{keyword}.*", "case_insensitive": True}}}
        for keyword in keywords
    ]
    bool_query = {"query": {"bool": {"must": match_queries}}}
    kwargs = {"sort": ["CiteScore:desc"]}
    results = await ds_client.search(
        settings.db.index_journals,
        bool_query,
        size=min(request.limit * 2, 10000),
        **kwargs,  # type: ignore
    )
    journals: list[JournalSuggestionResponse] = []

    unique_titles: set[str] = set()

    for result in results["hits"]["hits"]:
        source = result["_source"]
        if len(unique_titles) == request.limit:
            break
        if source["Title"] not in unique_titles:
            unique_titles.add(source["Title"])

            issn_types = ["E-ISSN", "Print ISSN"]
            issn_dict: dict[str, str | None] = {}
            for issn_type in issn_types:
                issn_dict[issn_type] = format_issn(source[issn_type])

            journals.append(
                JournalSuggestionResponse(
                    title=source["Title"],
                    citescore=source["CiteScore"],
                    eissn=issn_dict["E-ISSN"],
                    snip=source["SNIP"],
                    sjr=source["SJR"],
                    print_issn=issn_dict["Print ISSN"],
                )
            )

    if not journals:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.NO_DB_ENTRIES.value,
                "detail": "No journal found.",
            },
        )
    logger.info(f"Found journals in {time.time() - start}s.")
    return journals


@router.get(
    "/author",
    response_model=list[AuthorSuggestionResponse],
    responses={
        500: {
            "description": f"code={ErrorCode.NO_DB_ENTRIES.value} No author found.",
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                },
            },
        }
    },
)
async def author_suggestion(
    ds_client: Annotated[AsyncBaseSearch, Depends(get_ds_client)],
    settings: Annotated[Settings, Depends(get_settings)],
    request: AuthorSuggestionRequest = Depends(),  # noqa: B008
) -> list[AuthorSuggestionResponse]:
    """Suggest authors based on a name or part of a name.
    \f

    Parameters
    ----------
    request
        Body of the request sent to the server.
    ds_client
        Document store client
    settings
        Global settings of the application

    Returns
    -------
        A list of author names.
    """  # noqa: D301, D400, D205
    # Match case insensitive. Sadly aggregate queries don't support the CASE_INSENSITIVE flag
    # The /i flag is not supported either in regex queries.
    start = time.time()
    regex_pattern = "".join(
        [f"[{char.lower()}{char.upper()}]" for char in request.name]
    )

    pattern = re.compile(re.escape(request.name), re.IGNORECASE)

    # Regex query to partially match author name with a keyword field.
    query = {"regexp": {"authors": f".*{regex_pattern}.*"}}
    kwargs = {"_source": ["authors"]}
    # The size here is a bit arbitrary, but in theory it should ensure that
    # There is enough authors to fill the user's request. If not, people should
    # tell us right away.
    results = await ds_client.search(
        index=settings.db.index_paragraphs,
        query=query,
        size=request.limit * 10,
        aggs=None,
        timeout=60,
        **kwargs,
    )
    authors = []
    for res in results["hits"]["hits"]:
        for author in res["_source"]["authors"]:
            if author not in authors and pattern.search(author):
                authors.append(author)
                if len(authors) == request.limit:
                    break
        if len(authors) == request.limit:
            break

    authors = [AuthorSuggestionResponse(name=author) for author in authors]
    logger.info(f"Author suggestion took {time.time() - start}s")

    if not authors:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.NO_DB_ENTRIES.value,
                "detail": "No author found.",
            },
        )
    return authors


@router.get("/article_types", response_model=list[ArticleTypeSuggestionResponse])
async def article_types(
    ds_client: Annotated[AsyncBaseSearch, Depends(get_ds_client)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[ArticleTypeSuggestionResponse]:
    """List all available article_types in the db
    \f

    Parameters
    ----------
    ds_client
        Document store client
    settings
        Global settings of the application

    Returns
    -------
        A list of article types.
    """  # noqa: D301, D400, D205
    aggs = {"article_types": {"terms": {"field": "article_type", "size": 10000}}}
    results = await ds_client.search(
        index=settings.db.index_paragraphs, query=None, size=0, aggs=aggs
    )
    aggregation = results["aggregations"]["article_types"]["buckets"]
    unique_stripped_types = []
    unique_ids = []
    for i, article_type in enumerate(aggregation):
        stripped_type = article_type["key"].strip()
        if stripped_type not in unique_stripped_types:
            unique_stripped_types.append(stripped_type)
            unique_ids.append(i)
        else:
            agg_id = unique_ids[unique_stripped_types.index(stripped_type)]
            aggregation[agg_id]["doc_count"] += article_type["doc_count"]
    aggregation = [aggregation[i] for i in unique_ids]

    article_types = [
        ArticleTypeSuggestionResponse(
            article_type=article["key"], docs_in_db=article["doc_count"]
        )
        for article in aggregation
    ]
    logger.info(f"Found {len(article_types)} article types in the db.")
    return article_types



"""Custom middleware implementation."""

import datetime
import hashlib
import json
import logging
import time
from functools import cache
from typing import Any, Callable

from fastapi import HTTPException, Request, Response
from fastapi.responses import JSONResponse
from redis import Redis
from redis.asyncio import Redis as AsyncRedis
from redis.exceptions import ConnectionError as RedisConnectionError

from scholarag import __version__
from scholarag.app.config import Settings
from scholarag.app.dependencies import get_httpx_client, get_settings, get_user_id

logger = logging.getLogger(__name__)


@cache
def get_cache(settings: Settings) -> "AsyncRedis[Any] | None":
    """Load the cache client to cache requests."""
    if not settings.redis.host or not settings.redis.port:
        logger.warning("Redis url or port not set.")
        return None
    host = settings.redis.host
    if host.startswith("http://"):
        host = host[7:]
    r = Redis(host=host, port=settings.redis.port, decode_responses=True)
    r_async = AsyncRedis(host=host, port=settings.redis.port, decode_responses=True)
    try:
        # The connection goes through regardless, and an error is raised only after the first redis call.
        r.ping()
        logger.info("The caching service is enabled.")
    except RedisConnectionError:
        logger.info("The caching service is disabled.")
        r_async = None  # type: ignore
    return r_async


def select_relevant_settings(settings: Settings, path: str) -> list[str]:
    """Return relevant setting for hashing depending on endpoint selected."""
    # No need to deep copy it, since only one of the if/elif will be accessed.
    db_settings = [
        settings.db.db_type,
        settings.db.index_paragraphs,
        str(settings.db.index_journals),
        str(settings.db.user),
    ]
    if "/suggestions/" in path:
        # Take all db settings.
        relevant_settings_suggestions = db_settings
        return relevant_settings_suggestions

    elif "/qa/" in path:
        # Start with the DB settings.
        relevant_settings_qa = db_settings
        # Add retrieval settings.
        if "generative" in path:
            relevant_settings_qa.extend(
                [
                    str(settings.generative.openai.model),
                    str(settings.generative.openai.temperature),
                ]
            )
        relevant_settings_qa.append(str(settings.metadata.external_apis))

        return relevant_settings_qa

    elif "/retrieval/" in path:
        # Relevant for all retrieval endpoints.
        relevant_settings_retrieval = db_settings
        if path == "/retrieval/":
            # Add raw retrieval endpoint.
            relevant_settings_retrieval.append(str(settings.metadata.external_apis))
        return relevant_settings_retrieval

    else:
        # Ideally should never happen. Please append this function whenever a new enpoint is created.
        logger.warning(
            "Endpoint not recognized to generate the hash. Consider adding it to the"
            " function 'select_relevant_settings'."
        )
        return [settings.model_dump_json()]


async def custom_key_builder(request: Request, settings: Settings, version: str) -> str:
    """Build the unique caching key based on the request and the settings.

    Parameters
    ----------
    request
        Request sent by the user.
    settings
        Settings of the app at the moment of sending the request.
    version
        Version of the API.

    Returns
    -------
    key: Unique key used for caching.
    """
    relevant_settings = select_relevant_settings(settings, request.scope["path"])
    request_body = await request.body()
    cache_key = hashlib.md5(
        f"{version}:{request.query_params}:{request_body.decode('utf-8')}:{':'.join(relevant_settings)}".encode(),
        usedforsecurity=False,
    ).hexdigest()
    return f"{request.scope['path']}:{cache_key}"


async def strip_path_prefix(
    request: Request, call_next: Callable[[Any], Any]
) -> Response:
    """Optionally strip a prefix from a request path.

    Parameters
    ----------
    request
        Request sent by the user.
    call_next
        Function executed to get the output of the endpoint.

    Returns
    -------
    response: Response of the request after potentially stripping prefix from path and applying other middlewares
    """
    if request.base_url in (
        "http://testserver/",
        "http://test/",
    ) and "healthz" not in str(request.url):
        settings = request.app.dependency_overrides[get_settings]()
    else:
        settings = get_settings()
    prefix = settings.misc.application_prefix
    if prefix is not None and len(prefix) > 0 and request.url.path.startswith(prefix):
        new_path = request.url.path[len(prefix) :]
        scope = request.scope
        scope["path"] = new_path
        request = Request(scope, request.receive)
    return await call_next(request)


async def get_and_set_cache(
    request: Request, call_next: Callable[[Any], Any]
) -> Response:
    """Use caching for the requests.

    Parameters
    ----------
    request
        Request sent by the user.
    call_next
        Function executed to get the output of the endpoint.

    Returns
    -------
    response: Response of the request, potentially from the cache and with added headers.
    """
    # Specify here which endpoints should be cached.
    to_cache = [
        "/qa/generative",
        "/qa/streamed_generative",
        "/retrieval/article_count",
        "/retrieval/article_listing",
        "/retrieval/",
        "/suggestions/journal",
        "/suggestions/author",
    ]
    to_secure = [
        *to_cache,
        "/qa/passthrough",
        "/suggestions/article_types",
    ]
    if request.scope["path"] not in to_secure:
        response = await call_next(request)
        return response

    if request.base_url in ("http://testserver/", "http://test/"):
        settings = request.app.dependency_overrides[get_settings]()
    else:
        settings = get_settings()

    token = request.headers.get("Authorization")
    if token:
        token = token.rpartition(" ")[-1]
    httpx_client = await anext(get_httpx_client(settings))
    # If raises HTTPException return error as json.
    try:
        await get_user_id(token=token, settings=settings, httpx_client=httpx_client)
    except HTTPException as e:
        return JSONResponse(status_code=e.status_code, content=e.detail)

    if request.scope["path"] not in to_cache:
        response = await call_next(request)
        return response

    start = time.time()

    redis = get_cache(settings)

    # If no caching, run as usual.
    if not redis:
        response = await call_next(request)
        return response

    request_key = await custom_key_builder(
        request=request, settings=settings, version=__version__
    )

    # If the key exists in the db, return its result.
    if await redis.exists(request_key) and request.headers.get("cache-control") not in (
        "no-cache",
        "no-store",
    ):
        response = await redis.get(request_key)
        ttl = await redis.ttl(request_key)
        mapping = json.loads(response)
        mapping["content"] = json.dumps(mapping["content"])
        mapping.pop("state")
        mapping["headers"]["X-fastapi-cache"] = "Hit"
        mapping["headers"]["cache-control"] = f"max-age={ttl}s"
        mapping["headers"]["expires"] = (
            datetime.datetime.now() + datetime.timedelta(seconds=ttl)
        ).strftime(r"%d/%m/%Y, %H:%M:%S")
        mapping["headers"]["content-length"] = str(len(mapping["content"]))
        logger.info(
            f"Found cached request. cached request took {time.time() - start}s to"
            " complete."
        )
        return Response(**mapping)

    else:
        response = await call_next(request)
        if request.headers.get("cache-control") not in ("no-cache", "no-store"):
            # One cannot directly get the body, it is an async iterator.
            body = b""
            async for chunk in response.body_iterator:
                body += chunk

            # Prepare the response and add the headers.
            response = {
                "content": body.decode("utf-8"),
                "status_code": response.status_code,
                "headers": dict(response.headers),
                "media_type": response.media_type,
            }
            response["headers"]["X-fastapi-cache"] = "Miss"

            request_body = await request.body()
            cached = {
                **response,
                "state": {
                    "request_body": (
                        json.loads(request_body.decode("utf-8")) if request_body else ""
                    ),
                    "settings": json.loads(
                        settings.model_dump_json()
                    ),  # To avoid secrets being 'unjsonable'
                    "version": __version__,
                    "path": request.scope["path"],
                },
            }
            cached["content"] = cached["content"].split("<bbs_json_error>")[-1]
            cached["content"] = cached["content"].split("<bbs_json_data>")[-1]

            cached["content"] = json.loads(cached["content"])

            sec_time = datetime.timedelta(days=settings.redis.expiry)

            # Cache the response unless specified otherwise.
            await redis.set(request_key, json.dumps(cached), ex=sec_time)
            logger.info(f"Stored request in cache with key {request_key}")

            return Response(**response)
        else:
            return response



"""Run semantic search."""

import logging
from typing import Any

import pydantic.v1 as pydantic

from scholarag.document_stores import AsyncBaseSearch, BaseSearch

logger = logging.getLogger(__name__)


class RetrievalService(pydantic.BaseModel):
    """Class to run semantic search.

    Parameters
    ----------
    db_index_paragraphs
        Document store index for paragraphs.
    """

    db_index_paragraphs: str

    def run(
        self,
        ds_client: BaseSearch,
        query: str,
        retriever_k: int = 10,
        db_filter: dict[str, Any] | None = None,
        max_length: int = 100000,
    ) -> list[dict[str, Any]]:
        """Run bm25.

        Parameters
        ----------
        ds_client
            Documents store client.
        query
            Query to retrieve the most relevant documents from the documents store.
        retriever_k
            Number of documents to retrieve.
        db_filter
            Filtering parameters to add to the retrieval.
        max_length
            Maximum length of the documents to retrieve.

        Returns
        -------
        res: list[dict[str, Any]]
            List of retrieved documents saved under dictionary with all metadata.
        """
        res = ds_client.bm25_search(
            index_doc=self.db_index_paragraphs,
            query=query,
            filter_query=db_filter,
            k=retriever_k,
        )

        # Filtering long paragraphs
        res = [r for r in res if len(r["text"]) < max_length]
        return res

    async def arun(
        self,
        ds_client: AsyncBaseSearch,
        query: str,
        retriever_k: int = 10,
        db_filter: dict[str, Any] | None = None,
        max_length: int = 100000,
    ) -> list[dict[str, Any]]:
        """Run bm25.

        Parameters
        ----------
        ds_client
            Documents store client.
        query
            Query to retrieve the most relevant documents from the documents store.
        retriever_k
            Number of documents to retrieve.
        db_filter
            Filtering parameters to add to the retrieval.
        max_length
            Maximum length of the documents to retrieve.

        Returns
        -------
        res: list[dict[str, Any]]
            List of retrieved documents saved under dictionary with all metadata.
        """
        res = await ds_client.bm25_search(
            index_doc=self.db_index_paragraphs,
            query=query,
            filter_query=db_filter,
            k=retriever_k,
        )

        # Filtering long paragraphs
        res = [r for r in res if len(r["text"]) < max_length]
        return res



from unittest.mock import AsyncMock, Mock

import pytest
from openai import AsyncOpenAI, OpenAI
from openai.types.chat import ChatCompletion, ChatCompletionChunk, ChatCompletionMessage
from openai.types.chat.chat_completion import Choice as RunChoice
from openai.types.chat.chat_completion_chunk import Choice, ChoiceDelta
from openai.types.completion_usage import CompletionUsage
from scholarag.generative_question_answering import (
    SOURCES_SEPARATOR,
    GenerativeQAWithSources,
)


def test_run():
    fake_llm = Mock(spec=OpenAI(api_key="sasda"))
    gaq = GenerativeQAWithSources(client=fake_llm, model="gpt-ni.colas-turbo")
    query = "How nice is this context ?"
    context = [
        "This context is very nice.",
        "I really enjoyed this context.",
        "That's really an amazing context.",
    ]
    fake_response = ChatCompletion(
        id="chatcmpl-9k86ZK0uY65mPo1PvdnueBmD8hFLK",
        choices=[
            RunChoice(
                finish_reason="stop",
                index=0,
                logprobs=None,
                message=ChatCompletionMessage(
                    content=f"Very nice.\n{SOURCES_SEPARATOR}: 0",
                    role="assistant",
                    function_call=None,
                    tool_calls=None,
                ),
            )
        ],
        created=1720781271,
        model="bbp-3.5",
        object="chat.completion",
        service_tier=None,
        system_fingerprint=None,
        usage=CompletionUsage(
            completion_tokens=87, prompt_tokens=12236, total_tokens=12323
        ),
    )
    # Test with well formated output.
    fake_llm.chat.completions.create.return_value = fake_response
    result, finish_reason = gaq.run(query=query, contexts=context)
    assert result == {
        "answer": "Very nice.",
        "paragraphs": [0],
        "raw_answer": f"Very nice.\n{SOURCES_SEPARATOR}: 0",
    }
    assert finish_reason == "stop"
    # Test with a badly returned context that still makes sense.
    fake_response.choices[
        0
    ].message.content = f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice."
    result, finish_reason = gaq.run(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice.",
    }
    assert finish_reason == "stop"
    # Test with a badly formatted output but the context is in the source.
    fake_response.choices[0].message.content = (
        f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice.\nQUESTION: How"
        " nice is this context ?"
    )
    fake_response.choices[0].finish_reason = "length"
    result, finish_reason = gaq.run(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": (
            f"Very nice.\n{SOURCES_SEPARATOR}: This context is very"
            " nice.\nQUESTION: How nice is this context ?"
        ),
    }
    assert finish_reason == "length"
    # Test with a completely messed up output format
    fake_response.choices[
        0
    ].message.content = (
        "Very nice. This context is very nice.\nQUESTION: How nice is this context ?"
    )
    result, finish_reason = gaq.run(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": (
            "Very nice. This context is very nice.\nQUESTION: How nice is this"
            " context ?"
        ),
    }
    assert finish_reason == "length"
    # Test multiple sources. Correct format
    fake_response.choices[0].finish_reason = None
    fake_response.choices[0].message.content = f"Very nice.\n{SOURCES_SEPARATOR}: 0, 1"
    result, finish_reason = gaq.run(query=query, contexts=context)
    assert result == {
        "answer": "Very nice.",
        "paragraphs": [0, 1],
        "raw_answer": f"Very nice.\n{SOURCES_SEPARATOR}: 0, 1",
    }
    assert finish_reason is None

    # Test multiple sources. Sources returned not number
    fake_response.choices[0].message.content = (
        f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice., I really enjoyed"
        " this context."
    )
    result, finish_reason = gaq.run(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": (
            f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice., I really"
            " enjoyed this context."
        ),
    }

    # Test multiple sources. Wrong layout but sources still present.
    fake_response.choices[0].message.content = (
        f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice. I really enjoyed"
        " this context.\nQUESTION: How nice is this context ?"
    )
    result, finish_reason = gaq.run(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": (
            f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice. I really"
            " enjoyed this context.\nQUESTION: How nice is this context ?"
        ),
    }


@pytest.mark.asyncio
async def test_arun():
    fake_llm = AsyncMock(spec=AsyncOpenAI(api_key="assdas"))
    create_output = AsyncMock()
    gaq = GenerativeQAWithSources(client=fake_llm, model="gpt-ni.colas-turbo")
    query = "How nice is this context ?"
    context = [
        "This context is very nice.",
        "I really enjoyed this context.",
        "That's really an amazing context.",
    ]
    create_output.return_value = ChatCompletion(
        id="chatcmpl-9k86ZK0uY65mPo1PvdnueBmD8hFLK",
        choices=[
            RunChoice(
                finish_reason="stop",
                index=0,
                logprobs=None,
                message=ChatCompletionMessage(
                    content=f"Very nice.\n{SOURCES_SEPARATOR}: 0",
                    role="assistant",
                    function_call=None,
                    tool_calls=None,
                ),
            )
        ],
        created=1720781271,
        model="bbp-3.5",
        object="chat.completion",
        service_tier=None,
        system_fingerprint=None,
        usage=CompletionUsage(
            completion_tokens=87, prompt_tokens=12236, total_tokens=12323
        ),
    )
    # Test with well formated output.
    fake_llm.chat.completions.create = create_output
    result, finish_reason = await gaq.arun(query=query, contexts=context)
    assert result == {
        "answer": "Very nice.",
        "paragraphs": [0],
        "raw_answer": f"Very nice.\n{SOURCES_SEPARATOR}: 0",
    }
    assert finish_reason == "stop"
    # Test with a badly returned context that still makes sense.
    create_output.return_value.choices[
        0
    ].message.content = f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice."
    result, finish_reason = await gaq.arun(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice.",
    }
    assert finish_reason == "stop"
    # Test with a badly formatted output but the context is in the source.
    create_output.return_value.choices[0].message.content = (
        f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice.\nQUESTION: How"
        " nice is this context ?"
    )
    create_output.return_value.choices[0].finish_reason = "length"
    result, finish_reason = await gaq.arun(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": (
            f"Very nice.\n{SOURCES_SEPARATOR}: This context is very"
            " nice.\nQUESTION: How nice is this context ?"
        ),
    }
    assert finish_reason == "length"
    # Test with a completely messed up output format
    create_output.return_value.choices[
        0
    ].message.content = (
        "Very nice. This context is very nice.\nQUESTION: How nice is this context ?"
    )
    result, finish_reason = await gaq.arun(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": (
            "Very nice. This context is very nice.\nQUESTION: How nice is this"
            " context ?"
        ),
    }
    assert finish_reason == "length"
    # Test multiple sources. Correct format
    create_output.return_value.choices[0].finish_reason = None
    create_output.return_value.choices[
        0
    ].message.content = f"Very nice.\n{SOURCES_SEPARATOR}: 0, 1"
    result, finish_reason = await gaq.arun(query=query, contexts=context)
    assert result == {
        "answer": "Very nice.",
        "paragraphs": [0, 1],
        "raw_answer": f"Very nice.\n{SOURCES_SEPARATOR}: 0, 1",
    }
    assert finish_reason is None

    # Test multiple sources. Sources returned not number
    create_output.return_value.choices[0].message.content = (
        f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice., I really enjoyed"
        " this context."
    )
    result, finish_reason = await gaq.arun(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": (
            f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice., I really"
            " enjoyed this context."
        ),
    }

    # Test multiple sources. Wrong layout but sources still present.
    create_output.return_value.choices[0].message.content = (
        f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice. I really enjoyed"
        " this context.\nQUESTION: How nice is this context ?"
    )
    result, finish_reason = await gaq.arun(query=query, contexts=context)
    assert result == {
        "answer": None,
        "paragraphs": None,
        "raw_answer": (
            f"Very nice.\n{SOURCES_SEPARATOR}: This context is very nice. I really"
            " enjoyed this context.\nQUESTION: How nice is this context ?"
        ),
    }


def stream(**kwargs):
    base_response = ChatCompletionChunk(
        id="chatcmpl-9kBJNLoYybxyGe9pxsZkY3XgZMg8t",
        choices=[
            Choice(
                delta=ChoiceDelta(
                    content="", function_call=None, role="assistant", tool_calls=None
                ),
                finish_reason=None,
                index=0,
                logprobs=None,
            )
        ],
        created=1720793597,
        model="gpt-ni.colas-turbo",
        object="chat.completion.chunk",
        service_tier=None,
        system_fingerprint=None,
        usage=None,
    )
    to_stream = f"I am a great answer. {SOURCES_SEPARATOR}\n 0, 1"
    yield base_response
    for word in to_stream.split(" "):
        base_response.choices[0].delta.content = word + " " if word != "1" else word
        yield base_response
    yield ChatCompletionChunk(
        id="chatcmpl-9kBOdgQkrwK9yRbx7VXIiBXwTnKea",
        choices=[
            Choice(
                delta=ChoiceDelta(
                    content=None, function_call=None, role=None, tool_calls=None
                ),
                finish_reason="stop",
                index=0,
                logprobs=None,
            )
        ],
        created=1720793923,
        model="gpt-ni.colas-turbo",
        object="chat.completion.chunk",
        service_tier=None,
        system_fingerprint=None,
        usage=None,
    )
    yield ChatCompletionChunk(
        id="chatcmpl-9kBOdgQkrwK9yRbx7VXIiBXwTnKea",
        choices=[],
        created=1720793923,
        model="gpt-ni.ciolas-turbo",
        object="chat.completion.chunk",
        service_tier=None,
        system_fingerprint=None,
        usage=CompletionUsage(
            completion_tokens=15, prompt_tokens=3935, total_tokens=3950
        ),
    )


async def astream(**kwargs):
    base_response = ChatCompletionChunk(
        id="chatcmpl-9kBJNLoYybxyGe9pxsZkY3XgZMg8t",
        choices=[
            Choice(
                delta=ChoiceDelta(
                    content="", function_call=None, role="assistant", tool_calls=None
                ),
                finish_reason=None,
                index=0,
                logprobs=None,
            )
        ],
        created=1720793597,
        model="gpt-ni.colas-turbo",
        object="chat.completion.chunk",
        service_tier=None,
        system_fingerprint=None,
        usage=None,
    )
    to_stream = f"I am a great answer. {SOURCES_SEPARATOR}\n 0, 1"
    yield base_response
    for word in to_stream.split(" "):
        base_response.choices[0].delta.content = word + " " if word != "1" else word
        yield base_response
    yield ChatCompletionChunk(
        id="chatcmpl-9kBOdgQkrwK9yRbx7VXIiBXwTnKea",
        choices=[
            Choice(
                delta=ChoiceDelta(
                    content=None, function_call=None, role=None, tool_calls=None
                ),
                finish_reason="stop",
                index=0,
                logprobs=None,
            )
        ],
        created=1720793923,
        model="gpt-ni.colas-turbo",
        object="chat.completion.chunk",
        service_tier=None,
        system_fingerprint=None,
        usage=None,
    )
    yield ChatCompletionChunk(
        id="chatcmpl-9kBOdgQkrwK9yRbx7VXIiBXwTnKea",
        choices=[],
        created=1720793923,
        model="gpt-ni.ciolas-turbo",
        object="chat.completion.chunk",
        service_tier=None,
        system_fingerprint=None,
        usage=CompletionUsage(
            completion_tokens=15, prompt_tokens=3935, total_tokens=3950
        ),
    )


def test_stream():
    fake_llm = Mock(spec=OpenAI(api_key="assdas"))
    gaq = GenerativeQAWithSources(client=fake_llm, model="gpt-ni.colas-turbo")
    query = "How nice is this context ?"
    context = [
        "This context is very nice.",
        "I really enjoyed this context.",
        "That's really an amazing context.",
    ]
    fake_llm.chat.completions.create = stream
    streamed_gen = gaq.stream(query, context)
    try:
        partial_text = ""
        while True:
            partial_text += next(streamed_gen)
    except StopIteration as err:
        finish_reason = err.value
    assert partial_text == f"I am a great answer. {SOURCES_SEPARATOR}\n 0, 1"
    assert finish_reason == "stop"


@pytest.mark.asyncio
async def test_astream():
    fake_llm = AsyncMock(spec=AsyncOpenAI(api_key="assdas"))
    gaq = GenerativeQAWithSources(client=fake_llm, model="gpt-ni.colas-turbo")
    query = "How nice is this context ?"
    context = [
        "This context is very nice.",
        "I really enjoyed this context.",
        "That's really an amazing context.",
    ]
    fake_create = AsyncMock()
    fake_create.return_value = astream()
    fake_llm.chat.completions.create = fake_create
    try:
        partial_text = ""
        async for word in gaq.astream(query, context):
            partial_text += word
    except RuntimeError as err:
        finish_reason = err.args[0]
    assert partial_text == f"I am a great answer. {SOURCES_SEPARATOR}\n 0, 1"
    assert finish_reason == "stop"


@pytest.mark.parametrize(
    "raw_input, expected_answer, expected_paragraphs",
    [
        (f"Very nice.\n{SOURCES_SEPARATOR.upper()}: 0,1", "Very nice.", [0, 1]),
        (f"Very nice.{SOURCES_SEPARATOR.upper()}: 0,1", "Very nice.", [0, 1]),
        (f"Very nice.{SOURCES_SEPARATOR.lower()}: 0,1", "Very nice.", [0, 1]),
        (f"Very nice.\n{SOURCES_SEPARATOR.lower()}: 0,1", "Very nice.", [0, 1]),
        ("Wrongly formatted input.", None, None),
    ],
)
def test_process_raw_output(raw_input, expected_answer, expected_paragraphs):
    gaq = GenerativeQAWithSources(client=OpenAI(api_key="asaas"))
    res = gaq._process_raw_output(raw_input)
    assert res["answer"] == expected_answer
    assert res["paragraphs"] == expected_paragraphs



"""Parse articles and upload them on Elastic Search."""

from __future__ import annotations

import argparse
import asyncio
import json
import logging
import os
import pathlib
from typing import Any

from elasticsearch import ApiError
from elasticsearch.helpers import BulkIndexError as ESBulkIndexError
from opensearchpy.exceptions import TransportError
from opensearchpy.helpers import BulkIndexError as OSBulkIndexError

from scholarag.ds_utils import ds_upload, get_files, setup_parsing_ds

logger = logging.getLogger(__name__)


def get_parser() -> argparse.ArgumentParser:
    """Get parser for command line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "path",
        type=pathlib.Path,
        help="File or directory where are located the articles to parse.",
    )
    parser.add_argument(
        "parser_url",
        type=str,
        help="URL of the parser (needs to contain this address and the parser).",
    )
    parser.add_argument(
        "db_url",
        type=str,
        help="URL of the database instance (together with the port).",
    )
    parser.add_argument(
        "--articles-per-bulk",
        type=int,
        default=1000,
        help="Maximum number of articles to process before sending them to the db.",
    )
    parser.add_argument(
        "--max-paragraphs-length",
        "-u",
        type=int,
        help="Maximum length a paragraph is allowed to have to be uploaded to the DB.",
        default=None,
    )
    parser.add_argument(
        "--min-paragraphs-length",
        "-l",
        type=int,
        help="Minimum length a paragraph is allowed to have to be uploaded to the DB.",
        default=None,
    )
    parser.add_argument(
        "--multipart-params",
        default=None,
        type=json.loads,
        help=(
            "Dictionary for the multipart parameters for parsers that require some."
            " Must be passed as a json formated dict."
        ),
    )
    parser.add_argument(
        "--max-concurrent-requests",
        type=int,
        default=10,
        help=(
            "Maximum number of articles sent to the etl server PER BATCH OF"
            " 'articles-per-bulk' articles."
        ),
    )
    parser.add_argument(
        "--db-type",
        type=str,
        choices=["elasticsearch", "opensearch"],
        help="Type of database.",
        default="opensearch",
        required=False,
    )
    parser.add_argument(
        "--user",
        type=str,
        help=(
            "User of the database instance. The password is read from the environment"
            " under SCHOLARAG__DB__PASSWORD."
        ),
        required=False,
    )
    parser.add_argument(
        "--index",
        type=str,
        default="paragraphs",
        help="Desired name of the index holding paragraphs.",
    )
    parser.add_argument(
        "--files-failing-path",
        type=pathlib.Path,
        default=None,
        help="Path where to dump the files that failed. Expects a file.",
    )
    parser.add_argument(
        "-m",
        "--match-filename",
        type=str,
        help="""
        Parse only files with a name matching the given regular expression.
        Ignored when 'input_path' is a path to a file.
        """,
    )
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="""
        Parse files recursively.
        """,
    )
    parser.add_argument(
        "--use-ssl",
        action="store_true",
        help="Whether to verify ssl certificates or not.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Control verbosity",
    )
    return parser


async def run(
    path: pathlib.Path,
    recursive: bool,
    match_filename: str | None,
    db_url: str,
    parser_url: str,
    multipart_params: dict[str, Any] | None,
    max_concurrent_requests: int = 10,
    articles_per_bulk: int = 1000,
    min_paragraphs_length: int | None = None,
    max_paragraphs_length: int | None = None,
    db_type: str = "opensearch",
    index: str = "paragraphs",
    user: str | None = None,
    password: str | None = None,
    files_failing_path: pathlib.Path | None = None,
    use_ssl: bool = False,
) -> int:
    """Run the article parsing and upload results on ES."""
    ds_client, parsing_service = await setup_parsing_ds(
        db_url,
        db_type,
        user,
        password,
        max_concurrent_requests,
        use_ssl,
    )
    batches = get_files(
        path,
        recursive,
        match_filename=match_filename,
        articles_per_bulk=articles_per_bulk,
    )
    for i, batch in enumerate(batches):
        logger.info(f"Request server to parse batch {i}.")
        results = await parsing_service.arun(
            files=batch,
            url=parser_url,
            multipart_params=multipart_params,
        )
        logger.info("Collecting data to upload to the document store.")
        indices = [index for _ in range(len(batch))]
        try:
            files_failing = await ds_upload(
                filenames=batch,
                results=results,
                ds_client=ds_client,
                min_paragraphs_length=min_paragraphs_length,
                max_paragraphs_length=max_paragraphs_length,
                indices=indices,
            )
        except (ApiError, ESBulkIndexError, TransportError, OSBulkIndexError):
            files_failing = batch

    if files_failing and files_failing_path is not None:
        logger.info(f"Dumping failing files to {files_failing_path.resolve()}")
        with open(files_failing_path, "a") as f:
            for files in files_failing:
                f.write(json.dumps(files))
                f.write("\n")
    await ds_client.close()
    logger.info("Done.")

    return 0


def main() -> None:
    """Run the main logic."""
    # Parse command line
    parser = get_parser()
    args = parser.parse_args()

    # Additional parameters
    logging_level = logging.INFO if args.verbose else logging.WARNING

    # setup logging
    logging.basicConfig(
        format="[%(levelname)s]  %(asctime)s %(name)s  %(message)s", level=logging_level
    )
    password = os.getenv("SCHOLARAG__DB__PASSWORD")

    # Run the main function
    raise SystemExit(
        asyncio.run(
            run(
                path=args.path,
                recursive=args.recursive,
                match_filename=args.match_filename,
                db_url=args.db_url,
                db_type=args.db_type,
                parser_url=args.parser_url,
                multipart_params=args.multipart_params,
                max_concurrent_requests=args.max_concurrent_requests,
                articles_per_bulk=args.articles_per_bulk,
                min_paragraphs_length=args.min_paragraphs_length,
                max_paragraphs_length=args.max_paragraphs_length,
                index=args.index,
                user=args.user,
                password=password,
                files_failing_path=args.files_failing_path,
                use_ssl=args.use_ssl,
            )
        )
    )


if __name__ == "__main__":
    main()
